#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This file implements the compiler class for SAPI 5, used by
    Windows Speech Recognition (WSR).  This is the interface
    built into Windows Vista.
"""


#---------------------------------------------------------------------------

from win32com.client import constants

from ..base                     import CompilerBase, CompilerError
from ...grammar.rule_base       import Rule
from ...grammar.elements_basic  import Literal


#---------------------------------------------------------------------------
# Utility generator function for iterating over COM collections.

def collection_iter(collection):
    for index in range(0, collection.Count):
        yield collection.Item(index)

_trace_level=0
def trace_compile(func):
    return func
    def dec(self, element, src_state, dst_state, grammar, grammar_handle):
        global _trace_level
        s = '%s %s: compiling %s' % (grammar.name, '==='*_trace_level, element)
        l = 140-len(s)
        s += ' '*l + '| %-20s %s -> %s' % (src_state.Rule.Name, src_state and id(src_state), dst_state and id(dst_state))
        grammar._log_load.error(s)
        _trace_level+=1
        func(self, element, src_state, dst_state, grammar, grammar_handle)
        _trace_level-=1
        grammar._log_load.error('%s %s: compiling %s.' % (grammar.name, '...'*_trace_level, element))
    return dec


#---------------------------------------------------------------------------

class Sapi5Compiler(CompilerBase):

    #-----------------------------------------------------------------------
    # Methods for compiling grammars.

    def compile_grammar(self, grammar, context):
        self._log.debug("%s: Compiling grammar %s." % (self, grammar.name))
        grammar_handle = context.CreateGrammar()

        fake_rule = Rule(name="_FakeRule",
                         element=Literal("impossible " *20), exported=True)
        self._compile_rule(fake_rule, grammar, grammar_handle)
        for rule in grammar.rules:
            self._compile_rule(rule, grammar, grammar_handle)
        grammar_handle.Rules.Commit()

        return grammar_handle

    def _get_next_rule_id(self, grammar):
        # Generate a new unique rule ID within this grammar.
        if not hasattr(grammar, "_sapi5_next_rule_id"):
            grammar._sapi5_next_rule_id = 1
        rule_id = grammar._sapi5_next_rule_id
        grammar._sapi5_next_rule_id += 1
        return rule_id

    def _compile_rule(self, rule, grammar, grammar_handle):
        self._log.debug("%s: Compiling rule %s." % (self, rule.name))

        # Determine whether this rule has already been compiled.
        rule_handle = grammar_handle.Rules.FindRule(rule.name)
        if rule_handle:
            self._log.debug("%s: Already compiled rule %s." % (self, rule.name))
            return

        # Determine the flags to set when adding this rule.
        flags = 0
        if rule.exported:
            flags |= constants.SRATopLevel

        # Add this rule, and compile its root element.
        rule_id = self._get_next_rule_id(grammar)
        rule_handle = grammar_handle.Rules.Add(rule.name, flags, rule_id)
        self._log.debug("%s: Compiling rule %r (id %r)." % (self, rule_handle.Name, rule_handle.Id))
        self.compile_element(rule.element, rule_handle.InitialState, None, grammar, grammar_handle)

        return

        # Below is for debugging purposes only.
        stack = [(collection_iter(rule_handle.InitialState.Transitions), None)]
        while stack:
            self._log.error("%s: Stack len %s." % (self, len(stack)))
            try:
                t = next(stack[-1][0])
            except StopIteration:
                stack.pop()
                continue

            s = t.NextState
            if s:
                stack.append((collection_iter(s.Transitions), t))
            else:
                ts = [j for i,j in stack[1:]] + [t]
                path = [j.Text or j.Rule and ("<%s>" % j.Rule.Name) for j in ts]
                self._log.error("%s: path %r." % (self, path))
            if len(stack) > 100:
                ts = [j for i,j in stack[1:]] + [t]
                path = ["%r[%s]" % (j.Text,j.NextState) or j.Rule and ("<%s>" % j.Rule.Name) for j in ts]
                self._log.error("%s: path %r." % (self, path))
                break


    #-----------------------------------------------------------------------
    # Methods for compiling elements.

    @trace_compile
    def _compile_sequence(self, element, src_state, dst_state, grammar, grammar_handle):
        states = [src_state.Rule.AddState() for i in range(len(element.children)-1)]
        states.insert(0, src_state)
        states.append(dst_state)
        for i, child in enumerate(element.children):
            s1 = states[i]
            s2 = states[i + 1]
            self.compile_element(child, s1, s2, grammar, grammar_handle)

    @trace_compile
    def _compile_alternative(self, element, src_state, dst_state, grammar, grammar_handle):
        for child in element.children:
            self.compile_element(child, src_state, dst_state, grammar, grammar_handle)

    @trace_compile
    def _compile_optional(self, element, src_state, dst_state, grammar, grammar_handle):
        self.compile_element(element.children[0], src_state, dst_state, grammar, grammar_handle)
        src_state.AddWordTransition(dst_state, '')#None)

    @trace_compile
    def _compile_literal(self, element, src_state, dst_state, grammar, grammar_handle):
        src_state.AddWordTransition(dst_state, " ".join(element._words))

    @trace_compile
    def _compile_rule_ref(self, element, src_state, dst_state, grammar, grammar_handle):
        rule_handle = grammar_handle.Rules.FindRule(element.rule.name)
        if not rule_handle:
            grammar.add_rule(element.rule)
            self._compile_rule(element.rule, grammar, grammar_handle)
            rule_handle = grammar_handle.Rules.FindRule(element.rule.name)
            if not rule_handle:
                raise CompilerError("%s: Failed to create rule dependency: %r."
                                    % (self, element.rule.name))

        src_state.AddRuleTransition(dst_state, rule_handle)

    @trace_compile
    def _compile_list_ref(self, element, src_state, dst_state, grammar, grammar_handle):
        list_rule_name = "__list_%s" % element.list.name
        rule_handle = grammar_handle.Rules.FindRule(list_rule_name)
        if not rule_handle:
            grammar.add_list(element.list)
            flags = constants.SRADynamic
            rule_id = self._get_next_rule_id(grammar)
            rule_handle = grammar_handle.Rules.Add(list_rule_name, flags,
                                                   rule_id)
        src_state.AddRuleTransition(dst_state, rule_handle)

    @trace_compile
    def _compile_dictation(self, element, src_state, dst_state, grammar, grammar_handle):
        rule_handle = self._get_dictation_rule(grammar, grammar_handle)
        src_state.AddRuleTransition(dst_state, rule_handle)

    def _get_dictation_rule(self, grammar, grammar_handle):
        """
            Retrieve the special dictation rule.

            If it already exists within this grammar, return it.
            If it does not yet exist, create it.
        """
        rule_handle = grammar_handle.Rules.FindRule("dgndictation")
        if rule_handle:
            return rule_handle

        flags = 0
        # flags = constants.SRADynamic
        rule_id = self._get_next_rule_id(grammar)
        rule_handle = grammar_handle.Rules.Add("dgndictation", flags,
                                               rule_id)

        src_state = rule_handle.InitialState
        dst_state = None
        src_state.AddSpecialTransition(dst_state, 2)
        states = [src_state.Rule.AddState() for i in range(16)]
        src_state.AddSpecialTransition(states[0], 2)
        for s in states:
            s.AddSpecialTransition(dst_state, 2)
        for s1, s2 in zip(states[:-1], states[1:]):
            s1.AddSpecialTransition(s2, 2)

        return rule_handle

    @trace_compile
    def _compile_impossible(self, element, src_state, dst_state, grammar, grammar_handle):
        rule_handle = grammar_handle.Rules.FindRule("_FakeRule")
        src_state.AddRuleTransition(dst_state, rule_handle)

    @trace_compile
    def _compile_empty(self, element, src_state, dst_state, grammar, grammar_handle):
        src_state.AddWordTransition(dst_state, '')



#
# This file is part of Dragonfly.
# (c) Copyright 2019 by David Zurow
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Audio input/output classes for Kaldi backend
"""

from __future__ import division, print_function
import collections, contextlib, datetime, itertools, logging, os, time, threading, wave
from io import open

from six import PY2, binary_type, text_type, print_
from six.moves import queue, range
import sounddevice
import webrtcvad

from ..base import EngineError

_log = logging.getLogger("engine")


class MicAudio(object):
    """Streams raw audio from microphone. Data is received in a separate thread, and stored in a buffer, to be read from."""

    FORMAT = 'int16'
    SAMPLE_WIDTH = 2
    SAMPLE_RATE = 16000
    CHANNELS = 1
    BLOCKS_PER_SECOND = 100
    BLOCK_SIZE_SAMPLES = int(SAMPLE_RATE / float(BLOCKS_PER_SECOND))  # Block size in number of samples
    BLOCK_DURATION_MS = int(1000 * BLOCK_SIZE_SAMPLES // SAMPLE_RATE)  # Block duration in milliseconds

    def __init__(self, callback=None, buffer_s=0, flush_queue=True, start=True, input_device=None, self_threaded=None, reconnect_callback=None):
        self.callback = callback if callback is not None else lambda in_data: self.buffer_queue.put(in_data, block=False)
        self.flush_queue = bool(flush_queue)
        self.input_device = input_device
        self.self_threaded = bool(self_threaded)
        if reconnect_callback is not None and not callable(reconnect_callback):
            _log.error("Invalid reconnect_callback not callable: %r", reconnect_callback)
            reconnect_callback = None
        self.reconnect_callback = reconnect_callback

        self.buffer_queue = queue.Queue(maxsize=(buffer_s * 1000 // self.BLOCK_DURATION_MS))
        self.stream = None
        self.thread = None
        self.thread_cancelled = False
        self.device_info = None

        try:
            device_list = sounddevice.query_devices(device=self.input_device)
            if not device_list:
                raise EngineError("No audio devices found.")
        except ValueError as e:
            message = e.args[0]
            message += "\nAvailable devices are:\n" + str(sounddevice.query_devices())
            raise ValueError(message)

        self._connect(start=start)

    def _connect(self, start=None):
        callback = self.callback
        def proxy_callback(in_data, frame_count, time_info, status):
            callback(bytes(in_data))  # Must copy data from temporary C buffer!

        self.stream = sounddevice.RawInputStream(
            samplerate=self.SAMPLE_RATE,
            channels=self.CHANNELS,
            dtype=self.FORMAT,
            blocksize=self.BLOCK_SIZE_SAMPLES,
            # latency=80,
            device=self.input_device,
            callback=proxy_callback if not self.self_threaded else None,
        )

        if self.self_threaded:
            self.thread_cancelled = False
            self.thread = threading.Thread(target=self._reader_thread, args=(callback,))
            self.thread.daemon = True
            self.thread.start()

        if start:
            self.start()

        device_info = sounddevice.query_devices(self.stream.device)
        hostapi_info = sounddevice.query_hostapis(device_info['hostapi'])
        _log.info("streaming audio from '%s' using %s: %i sample_rate, %i block_duration_ms, %i latency_ms",
            device_info['name'], hostapi_info['name'], self.stream.samplerate, self.BLOCK_DURATION_MS, int(self.stream.latency*1000))
        self.device_info = device_info

    def _reader_thread(self, callback):
        while not self.thread_cancelled and self.stream and not self.stream.closed:
            if self.stream.active and self.stream.read_available >= self.stream.blocksize:
                in_data, overflowed = self.stream.read(self.stream.blocksize)
                # print('_reader_thread', read_available, len(in_data), overflowed, self.stream.blocksize)
                if overflowed:
                    _log.warning("audio stream overflow")
                callback(bytes(in_data))  # Must copy data from temporary C buffer!
            else:
                time.sleep(0.001)
                
    def _cancel_reader_thread(self):
        self.thread_cancelled = True
        if self.thread:
            self.thread.join()
            self.thread = None

    def destroy(self):
        self._cancel_reader_thread()
        if self.stream:
            self.stream.close()
            self.stream = None

    def reconnect(self):
        # FIXME: flapping
        old_device_info = self.device_info
        self._cancel_reader_thread()
        self.stream.close()
        self._connect(start=True)
        if self.reconnect_callback is not None:
            self.reconnect_callback(self)
        if self.device_info != old_device_info:
            raise EngineError("Audio reconnect could not reconnect to the same device")

    def start(self):
        self.stream.start()

    def stop(self):
        self.stream.stop()

    def read(self, nowait=False):
        """Return a block of audio data. If nowait==False, waits for a block if necessary; else, returns False immediately if no block is available."""
        if self.stream or (self.flush_queue and not self.buffer_queue.empty()):
            if nowait:
                try:
                    return self.buffer_queue.get_nowait()  # Return good block if available
                except queue.Empty as e:
                    return False  # Queue is empty for now
            else:
                return self.buffer_queue.get()  # Wait for a good block and return it
        else:
            return None  # We are done

    def read_loop(self, callback):
        """Block looping reading, repeatedly passing a block of audio data to callback."""
        for block in iter(self):
            callback(block)

    def iter(self, nowait=False):
        """Generator that yields all audio blocks from microphone."""
        while True:
            block = self.read(nowait=nowait)
            if block is None:
                break
            yield block

    def __iter__(self):
        """Generator that yields all audio blocks from microphone."""
        return self.iter()

    def get_wav_length_s(self, data):
        assert isinstance(data, binary_type)
        length_bytes = len(data)
        assert self.FORMAT == 'int16'
        length_samples = length_bytes / self.SAMPLE_WIDTH
        return (float(length_samples) / self.SAMPLE_RATE)

    def write_wav(self, filename, data):
        # _log.debug("write wav %s", filename)
        wf = wave.open(filename, 'wb')
        wf.setnchannels(self.CHANNELS)
        # wf.setsampwidth(self.pa.get_sample_size(FORMAT))
        assert self.FORMAT == 'int16'
        wf.setsampwidth(self.SAMPLE_WIDTH)
        wf.setframerate(self.SAMPLE_RATE)
        wf.writeframes(data)
        wf.close()

    @staticmethod
    def print_list():
        print_("")
        print_("LISTING ALL INPUT DEVICES SUPPORTED BY PORTAUDIO")
        print_("(any device numbers not shown are for output only)")
        print_("")
        devices = sounddevice.query_devices()
        print_(devices)

        # for i in range(0, pa.get_device_count()):
        #     info = pa.get_device_info_by_index(i)

        #     if info['maxInputChannels'] > 0:  # microphone? or just speakers
        #         print_("DEVICE #%d" % info['index'])
        #         print_("    %s" % info['name'])
        #         print_("    input channels = %d, output channels = %d, defaultSampleRate = %d" %
        #             (info['maxInputChannels'], info['maxOutputChannels'], info['defaultSampleRate']))
        #         # print_(info)
        #         try:
        #           supports16k = pa.is_format_supported(16000,  # sample rate
        #               input_device = info['index'],
        #               input_channels = info['maxInputChannels'],
        #               input_format = pyaudio.paInt16)
        #         except ValueError:
        #           print_("    NOTE: 16k sampling not supported, configure pulseaudio to use this device")

        print_("")


class VADAudio(MicAudio):
    """Filter & segment audio with voice activity detection."""

    def __init__(self, aggressiveness=3, **kwargs):
        super(VADAudio, self).__init__(**kwargs)
        self.vad = webrtcvad.Vad(aggressiveness)

    def vad_collector(self, start_window_ms=150, start_padding_ms=100,
        end_window_ms=150, end_padding_ms=None, complex_end_window_ms=None,
        ratio=0.8, blocks=None, nowait=False, audio_auto_reconnect=False,
        ):
        """Generator/coroutine that yields series of consecutive audio blocks comprising each phrase, separated by yielding a single None.
            Determines voice activity by ratio of blocks in window_ms. Uses a buffer to include window_ms prior to being triggered.
            Example: (block, ..., block, None, block, ..., block, None, ...)
                      |----phrase-----|        |----phrase-----|
        """
        assert end_padding_ms == None, "end_padding_ms not supported yet"
        num_start_window_blocks = max(1, int(start_window_ms // self.BLOCK_DURATION_MS))
        num_start_padding_blocks = max(0, int((start_padding_ms or 0) // self.BLOCK_DURATION_MS))
        num_end_window_blocks = max(1, int(end_window_ms // self.BLOCK_DURATION_MS))
        num_complex_end_window_blocks = max(1, int((complex_end_window_ms or end_window_ms) // self.BLOCK_DURATION_MS))
        num_end_padding_blocks = max(0, int((end_padding_ms or 0) // self.BLOCK_DURATION_MS))
        _log.debug("%s: vad_collector: num_start_window_blocks=%s num_end_window_blocks=%s num_complex_end_window_blocks=%s",
            self, num_start_window_blocks, num_end_window_blocks, num_complex_end_window_blocks)
        audio_reconnect_threshold_blocks = 5
        audio_reconnect_threshold_time = 50 * self.BLOCK_DURATION_MS / 1000

        ring_buffer = collections.deque(maxlen=max(
            (num_start_window_blocks + num_start_padding_blocks),
            (num_end_window_blocks + num_end_padding_blocks),
            (num_complex_end_window_blocks + num_end_padding_blocks),
        ))
        ring_buffer_recent_slice = lambda num_blocks: itertools.islice(ring_buffer, max(0, (len(ring_buffer) - num_blocks)), None)

        triggered = False
        in_complex_phrase = False
        num_empty_blocks = 0
        last_good_block_time = time.time()

        if blocks is None: blocks = self.iter(nowait=nowait)
        for block in blocks:
            if block is False or block is None:
                # Bad/empty block
                num_empty_blocks += 1
                if audio_auto_reconnect and (num_empty_blocks >= audio_reconnect_threshold_blocks) and (time.time() - last_good_block_time >= audio_reconnect_threshold_time):
                    _log.warning("%s: no good block received recently, so reconnecting audio", self)
                    self.reconnect()
                    num_empty_blocks = 0
                    last_good_block_time = time.time()
                in_complex_phrase = yield False

            else:
                # Good block
                num_empty_blocks = 0
                last_good_block_time = time.time()
                is_speech = self.vad.is_speech(block, self.SAMPLE_RATE)

                if not triggered:
                    # Between phrases
                    ring_buffer.append((block, is_speech))
                    num_voiced = len([1 for (_, speech) in ring_buffer_recent_slice(num_start_window_blocks) if speech])
                    if num_voiced >= (num_start_window_blocks * ratio):
                        # Start of phrase
                        triggered = True
                        for block, _ in ring_buffer_recent_slice(num_start_padding_blocks + num_start_window_blocks):
                            # print('|' if is_speech else '.', end='')
                            # print('|' if in_complex_phrase else '.', end='')
                            in_complex_phrase = yield block
                        # print('#', end='')
                        ring_buffer.clear()

                else:
                    # Ongoing phrase
                    in_complex_phrase = yield block
                    # print('|' if is_speech else '.', end='')
                    # print('|' if in_complex_phrase else '.', end='')
                    ring_buffer.append((block, is_speech))
                    num_unvoiced = len([1 for (_, speech) in ring_buffer_recent_slice(num_end_window_blocks) if not speech])
                    num_complex_unvoiced = len([1 for (_, speech) in ring_buffer_recent_slice(num_complex_end_window_blocks) if not speech])
                    if (not in_complex_phrase and num_unvoiced >= (num_end_window_blocks * ratio)) or \
                        (in_complex_phrase and num_complex_unvoiced >= (num_complex_end_window_blocks * ratio)):
                        # End of phrase
                        triggered = False
                        in_complex_phrase = yield None
                        # print('*')
                        ring_buffer.clear()

        if triggered:
            # We were in a phrase, so we must terminate it (this may be abrupt!)
            yield None

    def debug_print_simple(self):
        print("block_duration_ms=%s" % self.BLOCK_DURATION_MS)
        for block in self.iter(nowait=False):
            is_speech = self.vad.is_speech(block, self.SAMPLE_RATE)
            print('|' if is_speech else '.', end='')

    def debug_loop(self, *args, **kwargs):
        audio_iter = self.vad_collector(*args, **kwargs)
        next(audio_iter)
        while True:
            block = audio_iter.send(False)


class AudioStore(object):
    """
    Stores the current audio data being recognized, which is cleared upon calling `finalize()`.
    Also, optionally stores the last `maxlen` recognitions as `AudioStoreEntry` objects,
    indexed in reverse order (0 is most recent), and advanced upon calling `finalize()`.
    Note: `finalize()` should be called after the recognition has been parsed and its actions executed.

    Constructor arguments:
    - *maxlen* (*int*, default *None*): if set, the number of previous recognitions to temporarily store.
    - *save_dir* (*str*, default *None*): if set, the directory to save the `retain.tsv` file and optionally wav files.
    - *save_metadata* (*bool*, default *None*): whether to automatically save the recognition metadata.
    - *save_audio* (*bool*, default *None*): whether to automatically save the recognition audio data (in addition to just the recognition metadata).
    - *retain_approval_func* (*Callable*, default *None*): if set, will be called with the `AudioStoreEntry` object about to be saved,
        and should return `bool` whether to actually save. Example: `retain_approval_func=lambda entry: bool(entry.grammar_name != 'noisegrammar')`
    """

    def __init__(self, audio_obj, maxlen=None, save_dir=None, save_audio=None, save_metadata=None, retain_approval_func=None):
        self.audio_obj = audio_obj
        self.maxlen = maxlen
        self.save_dir = save_dir
        self.save_audio = save_audio
        self.save_metadata = save_metadata
        if self.save_dir:
            _log.info("retaining recognition audio and/or metadata to '%s'", self.save_dir)
        self.retain_approval_func = retain_approval_func
        self.deque = collections.deque(maxlen=maxlen) if maxlen else None
        self.blocks = []

    current_audio_data = property(lambda self: b''.join(bytes(self.blocks)) if PY2 else b''.join(self.blocks))
    current_audio_length_ms = property(lambda self: len(self.blocks) * self.audio_obj.BLOCK_DURATION_MS)

    def add_block(self, block):
        self.blocks.append(block)

    def finalize(self, text, grammar_name, rule_name, likelihood=None, tag='', has_dictation=None):
        """ Finalizes current utterance, creating its AudioStoreEntry and saving it (if enabled). """
        entry = AudioStoreEntry(self.current_audio_data, grammar_name, rule_name, text, likelihood, tag, has_dictation)
        if self.deque is not None:
            if len(self.deque) == self.deque.maxlen:
                self.save(-1)  # Save oldest, which is about to be evicted
            self.deque.appendleft(entry)
        self.blocks = []

    def cancel(self):
        self.blocks = []

    def save(self, index):
        """ Saves AudioStoreEntry for given index (0 is most recent). """
        if slice(index).indices(len(self.deque))[1] >= len(self.deque):
            raise EngineError("Invalid index to save in AudioStore")
        if not self.save_dir:
            return
        if not os.path.isdir(self.save_dir):
            _log.warning("Recognition data was not retained because '%s' was not a directory" % self.save_dir)
            return

        entry = self.deque[index]
        if (not self.save_audio) and (not self.save_metadata) and (not entry.force_save):
            return
        if self.retain_approval_func and not self.retain_approval_func(entry):
            return
        if self.save_audio or entry.force_save:
            filename = os.path.join(self.save_dir, "retain_%s.wav" % datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S_%f"))
            self.audio_obj.write_wav(filename, entry.audio_data)
        else:
            filename = ''

        with open(os.path.join(self.save_dir, "retain.tsv"), 'a', encoding='utf-8') as tsv_file:
            tsv_file.write(u'\t'.join([
                    filename,
                    text_type(self.audio_obj.get_wav_length_s(entry.audio_data)),
                    entry.grammar_name,
                    entry.rule_name,
                    entry.text,
                    text_type(entry.likelihood),
                    text_type(entry.tag),
                    text_type(entry.has_dictation),
                ]) + '\n')

    def save_all(self, remove=True):
        if self.deque:
            for i in reversed(range(len(self.deque))):
                self.save(i)
            if remove:
                self.deque.clear()

    def __getitem__(self, key):
        return self.deque[key]
    def __len__(self):
        return len(self.deque)
    def __bool__(self):
        return True
    __nonzero__ = __bool__  # PY2 compatibility

class AudioStoreEntry(object):
    __slots__ = ('audio_data', 'grammar_name', 'rule_name', 'text', 'likelihood', 'tag', 'has_dictation', 'force_save')

    def __init__(self, audio_data, grammar_name, rule_name, text, likelihood, tag, has_dictation, force_save=False):
        self.audio_data = audio_data
        self.grammar_name = grammar_name
        self.rule_name = rule_name
        self.text = text
        self.likelihood = likelihood
        self.tag = tag
        self.has_dictation = has_dictation
        self.force_save = force_save

    def set(self, key, value):
        """ Sets given key (as *str*) to value, returning the AudioStoreEntry for chaining; usable in lambda functions. """
        setattr(self, key, value)
        return self


class WavAudio(object):
    """ Class for mimicking normal microphone input, but from wav files. """

    @classmethod
    def read_file(cls, filename, realtime=False):
        """ Yields raw audio blocks from wav file, terminated by a None element. """
        if not os.path.isfile(filename):
            raise IOError("'%s' is not a file. Please use a different file path.")

        with contextlib.closing(wave.open(filename, 'rb')) as file:
            # Validate the wave file's header
            if file.getnchannels() != MicAudio.CHANNELS:
                raise ValueError("WAV file '%s' should use %d channel(s), not %d!"
                           % (filename, MicAudio.CHANNELS, file.getnchannels()))
            elif file.getsampwidth() != MicAudio.SAMPLE_WIDTH:
                raise ValueError("WAV file '%s' should use sample width %d, not "
                           "%d!" % (filename, MicAudio.SAMPLE_WIDTH, file.getsampwidth()))
            elif file.getframerate() != MicAudio.SAMPLE_RATE:
                raise ValueError("WAV file '%s' should use sample rate %d, not "
                           "%d!" % (filename, MicAudio.SAMPLE_RATE, file.getframerate()))

            next_time = time.time()
            for _ in range(0, int(file.getnframes() / MicAudio.BLOCK_SIZE_SAMPLES) + 1):
                data = file.readframes(MicAudio.BLOCK_SIZE_SAMPLES)
                if not data:
                    break
                if realtime:
                    time_behind = next_time - time.time()
                    if time_behind > 0:
                        time.sleep(time_behind)
                    next_time += float(MicAudio.BLOCK_SIZE_SAMPLES) / MicAudio.SAMPLE_RATE
                yield data
            yield None

    @classmethod
    def read_file_with_vad(cls, filename, realtime=False, **kwargs):
        """ Yields raw audio blocks from wav file, after processing by VAD, terminated by a None element. """
        vad_audio = VADAudio()
        vad_audio_iter = vad_audio.vad_collector(blocks=cls.read_file(filename, realtime=realtime), **kwargs)
        return vad_audio_iter



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
"""

import unittest
from .rule_test_grammar     import RuleTestGrammar
from ..engines              import get_engine


#===========================================================================

class RuleTestCase(unittest.TestCase):
    """
    """

    def run(self, result=None):
        self.engine = get_engine()
        self.grammar = RuleTestGrammar()
        return unittest.TestCase.run(self, result)

    def tearDown(self):
        self.grammar.unload()
        for rule in self.grammar.rules:
            self.grammar.remove_rule(rule)
        for lst in self.grammar.lists:
            self.grammar.remove_list(lst)

    def add_rule(self, rule):
        self.grammar.add_rule(rule)

    def recognize(self, words):
        return self.grammar.recognize(words)

    def recognize_node(self, words):
        return self.grammar.recognize_node(words)

    def recognize_extras(self, words):
        return self.grammar.recognize_extras(words)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This example command module demonstrates how to control the engine from
    grammars.

    This module can be activated in the same way as other
    Natlink macros by placing it in the "NatLink/MacroSystem" folder.

"""

import sys

from dragonfly import Grammar, MappingRule, get_engine, Function


#---------------------------------------------------------------------------
# Create this module's grammar.

grammar = Grammar("Engine control")


#---------------------------------------------------------------------------
# Create a mapping rule which maps things you can say to actions.

def disconnect():
    print("Disconnecting engine.")
    engine = get_engine()
    if engine.name == 'natlink' and 'natspeak' in sys.executable:
        print("Not calling disconnect() for embedded natlink.")
    else:
        engine.disconnect()


control_rule = MappingRule(
    name="control",
    mapping={
        # Stop recognising from the microphone and exit.
        "disconnect engine|stop recognizing": Function(disconnect),
    }
)

# Add the action rule to the grammar instance.
grammar.add_rule(control_rule)


#---------------------------------------------------------------------------
# Load the grammar instance and define how to unload it.

grammar.load()

# Unload function which will be called by natlink at unload time.
def unload():
    global grammar
    if grammar: grammar.unload()
    grammar = None



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import unittest

from six import text_type, string_types

from dragonfly.engines import get_engine, EngineBase
from dragonfly.engines.base.dictation import DictationContainerBase

#---------------------------------------------------------------------------


class TestEngineSapi5(unittest.TestCase):

    def test_get_engine_sapi5_is_usable(self):
        """ Verify that the sapi5 engine is usable. """
        engine = get_engine()
        self.assertTrue(isinstance(engine, EngineBase))
        self.assertTrue(engine.name.startswith("sapi5"))

        engine.speak("testing WSR")
        from dragonfly import Literal, Sequence
        from dragonfly.test import ElementTester
        seq = Sequence([Literal("hello"), Literal("world")])
        tester = ElementTester(seq, engine=engine)
        results = tester.recognize("hello world")
        self.assertEqual([u"hello", u"world"], results)

    def test_dictation(self):
        # Test dictation separately for SAPI5 because test_dictation.py
        # won't work with it.
        from dragonfly import Dictation, Literal, Sequence
        from dragonfly.test import ElementTester, RecognitionFailure
        seq = Sequence([Literal("hello"), Dictation("text")])
        tester = ElementTester(seq)

        # Test one word.
        results = tester.recognize("hello world")
        assert results[0] == "hello"

        # Verify recognition returned dictation result.
        dictation = results[1]
        if not isinstance(dictation, DictationContainerBase):
            message = (u"Expected recognition result to be a dictation"
                       u" container, but received %r"
                       % (repr(dictation).decode("windows-1252"),))
            self.fail(message.encode("windows-1252"))

        # Verifying dictation converts/encode successfully.
        self.assertEqual(str(dictation), "world")
        self.assertEqual(text_type(dictation), "world")
        self.assertTrue(isinstance(repr(dictation), string_types))

        # Test incomplete.
        results = tester.recognize("hello")
        assert results is RecognitionFailure

    def test_recognition_observers(self):
        # RecognitionObservers are a bit quirky for the sapi5 engines,
        # so the tests for them are repeated here to handle that.
        from dragonfly import (Integer, Literal, RecognitionHistory,
                               RecognitionObserver)
        from dragonfly.test import ElementTester, RecognitionFailure

        class RecognitionObserverTester(RecognitionObserver):
            """ RecognitionObserver class from the recobs doctests. """

            def __init__(self):
                RecognitionObserver.__init__(self)
                self.waiting = False
                self.words = None

            def on_begin(self):
                self.waiting = True

            def on_recognition(self, words):
                self.waiting = False
                self.words = words

            def on_failure(self):
                self.waiting = False
                self.words = False

        test_recobs = RecognitionObserverTester()
        test_recobs.register()
        results = test_recobs.waiting, test_recobs.words
        assert results == (False, None)

        # Test simple literal element recognitions.
        test_lit = ElementTester(Literal("hello world"))
        assert test_lit.recognize("hello world") == "hello world"
        results = test_recobs.waiting, test_recobs.words
        assert results == (False, (u'hello', u'world'))
        assert test_lit.recognize("hello universe") is RecognitionFailure
        results = test_recobs.waiting, test_recobs.words
        assert results == (False, False)

        # Test Integer element recognitions
        test_int = ElementTester(Integer(min=1, max=100))
        assert test_int.recognize("seven") == 7
        results = test_recobs.waiting, test_recobs.words
        assert results == (False, (u'seven',))
        assert test_int.recognize("forty seven") == 47
        results = test_recobs.waiting, test_recobs.words
        assert results == (False, (u'forty', u'seven'))
        assert test_int.recognize("one hundred") is RecognitionFailure
        results = test_recobs.waiting, test_recobs.words
        assert results == (False, False)
        assert test_lit.recognize("hello world") == u'hello world'

        # Now test RecognitionHistory.
        history = RecognitionHistory()
        assert test_lit.recognize("hello world") == u'hello world'

        # Not yet registered, so didn't receive previous recognition.
        assert history == []
        history.register()
        assert test_lit.recognize("hello world") == u'hello world'

        # Now registered, so should have received previous recognition.
        assert history == [(u'hello', u'world')]
        assert test_lit.recognize("hello universe") is RecognitionFailure

        # Failed recognitions are ignored, so history is unchanged.
        assert history == [(u'hello', u'world')]
        assert test_int.recognize("eighty six") == 86
        assert history == [(u'hello', u'world'), (u'eighty', u'six')]

        # The RecognitionHistory class allows its maximum length to be set.
        history = RecognitionHistory(3)
        history.register()
        assert history == []
        for i, word in enumerate(["one", "two", "three", "four", "five"]):
            assert test_int.recognize(word) == i + 1
        assert history == [(u'three',), (u'four',), (u'five',)]

        history = RecognitionHistory(1)
        history.register()
        assert history == []
        for i, word in enumerate(["one", "two", "three", "four", "five"]):
            assert test_int.recognize(word) == i + 1
        assert history == [(u'five',)]



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# pylint: disable=global-statement

import logging
import os

from .base import EngineBase, EngineError, MimicFailure


# ---------------------------------------------------------------------------

_default_engine = None
_engines_by_name = {}


def get_engine(name=None, **kwargs):
    """
        Get the engine implementation.

        This function will initialize an engine instance using the
        ``get_engine`` and ``is_engine_available`` functions in the engine
        packages and return an instance of the first available engine.  If
        one has already been initialized, it will be returned instead.

        If no specific engine is requested and no engine has already been
        initialized, this function will initialize and return an instance of
        the first available engine in the following order:

         =======================   =========================================
         SR engine back-end        Engine name string(s)
         =======================   =========================================
         1. Dragon/Natlink         ``"natlink"``
         2. Kaldi                  ``"kaldi"``
         3. WSR/SAPI 5             ``"sapi5", "sapi5inproc", "sapi5shared"``
         4. CMU Pocket Sphinx      ``"sphinx"``
         =======================   =========================================

        The :ref:`Text-input engine <RefTextEngine>` can be initialized by
        specifying ``"text"`` as the engine name.  This back-end will
        **not** be initialized if no specific engine is requested because
        the back-end is not a real SR engine and is used mostly for testing.

        **Arguments**:

        :param name: optional human-readable name of the engine to return.
        :type name: str
        :param \\**kwargs: optional keyword arguments passed through to the
            engine for engine-specific configuration.
        :rtype: EngineBase
        :returns: engine instance
        :raises: EngineError
    """
    # pylint: disable=too-many-statements,too-many-branches
    global _default_engine, _engines_by_name
    log = logging.getLogger("engine")

    if name and name in _engines_by_name:
        # If the requested engine has already been initialized, return it.
        engine = _engines_by_name[name]
    elif not name and _default_engine:
        # If no specific engine is requested and an engine has already
        #  been initialized, return it.
        engine = _default_engine
    else:
        # No engine has been initialized yet.
        engine = None

    # Check if there is an already initialized engine *and* custom engine
    #  initialization arguments.  This is not allowed.
    if engine and kwargs is not None and len(kwargs) > 0:
        message = ("Error: Passing get_engine arguments to an engine "
                   "that has already been initialized, hence these "
                   "arguments are ignored.")
        log.error(message)
        raise EngineError(message)

    # If there is a relevant initialized engine already, then return it.
    if engine:
        return engine

    # Check if we're on Windows. If  we're not on Windows, then we don't
    #  evaluate Windows-only engines like natlink.
    windows = os.name == 'nt'

    if not engine and windows and name in (None, "natlink"):
        # Attempt to retrieve the natlink back-end.
        try:
            from .backend_natlink import is_engine_available
            from .backend_natlink import get_engine as get_specific_engine
            if is_engine_available(**kwargs):
                engine = get_specific_engine(**kwargs)
        except Exception as e:
            message = ("Exception while initializing natlink engine:"
                       " %s" % (e,))
            log.warning(message)
            if name:
                raise EngineError(message)

    if not engine and name in (None, "kaldi"):
        # Attempt to retrieve the Kaldi back-end.
        try:
            from .backend_kaldi import is_engine_available
            from .backend_kaldi import get_engine as get_specific_engine
            if is_engine_available(**kwargs):
                engine = get_specific_engine(**kwargs)
        except Exception as e:
            message = ("Exception while initializing kaldi engine:"
                       " %s" % (e,))
            log.warning(message)
            if name:
                raise EngineError(message)

    sapi5_names = (None, "sapi5shared", "sapi5inproc", "sapi5")
    if not engine and windows and name in sapi5_names:
        # Attempt to retrieve the sapi5 back-end.
        try:
            from .backend_sapi5 import is_engine_available
            from .backend_sapi5 import get_engine as get_specific_engine
            if is_engine_available(name, **kwargs):
                engine = get_specific_engine(name, **kwargs)
        except Exception as e:
            message = ("Exception while initializing sapi5 engine:"
                       " %s" % (e,))
            log.warning(message)
            if name:
                raise EngineError(message)

    if not engine and name in (None, "sphinx"):
        # Attempt to retrieve the CMU Sphinx back-end.
        try:
            from .backend_sphinx import is_engine_available
            from .backend_sphinx import get_engine as get_specific_engine
            if is_engine_available(**kwargs):
                engine = get_specific_engine(**kwargs)
        except Exception as e:
            message = ("Exception while initializing sphinx engine:"
                       " %s" % (e,))
            log.warning(message)
            if name:
                raise EngineError(message)

    # Only retrieve the text input engine if explicitly specified; it is not
    #  an actual SR engine implementation and is mostly intended to be used
    #  for testing.
    if not engine and name == "text":
        # Attempt to retrieve the TextInput engine instance.
        try:
            from .backend_text import is_engine_available
            from .backend_text import get_engine as get_specific_engine
            if is_engine_available(**kwargs):
                engine = get_specific_engine(**kwargs)
        except Exception as e:
            message = ("Exception while initializing text-input engine:"
                       " %s" % (e,))
            log.warning(message)
            if name:
                raise EngineError(message)

    # Return the engine instance, if one has been initialized.  Log a
    #  message about which SR engine back-end was used.
    if engine:
        message = "Initialized %r SR engine: %r." % (engine.name, engine)
        log.info(message)
        return engine
    elif not name:
        raise EngineError("No usable engines found.")
    else:
        valid_names = ["natlink", "kaldi", "sphinx", "sapi5shared",
                       "sapi5inproc", "sapi5", "text"]
        if name not in valid_names:
            raise EngineError("Requested engine %r is not a valid engine "
                              "name." % (name,))
        else:
            raise EngineError("Requested engine %r not available."
                              % (name,))


def get_current_engine():
    """
        Get the currently initialized SR engine object.

        If an SR engine has not been initialized yet, ``None`` will be
        returned instead.

        :rtype: EngineBase | None
        :returns: engine object or None

        Usage example:

        .. code-block:: python

           # Print the name of the current engine if one has been
           # initialized.
           from dragonfly import get_current_engine
           engine = get_current_engine()
           if engine:
               print("Engine name: %r" % engine.name)
           else:
               print("No engine has been initialized.")

    """
    global _default_engine
    return _default_engine


# ---------------------------------------------------------------------------

def register_engine_init(engine):
    """
        Register initialization of an engine.

        This function sets the default engine to the first engine
        initialized.

    """

    global _default_engine, _engines_by_name
    if not _default_engine:
        _default_engine = engine
    if engine and engine.name not in _engines_by_name:
        _engines_by_name[engine.name] = engine



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import sys
import unittest
from six import text_type

from dragonfly        import *
from ..test           import TestError, RecognitionFailure
from .element_tester  import ElementTester


#===========================================================================

class ElementTestCase(unittest.TestCase):
    """
        Test case class for easy testing of language elements.

        This class is usually used in one of the following two ways:
        #. Set the class attribute :attr:`_element`.
        #. Override the :meth:`_build_element` method.  By default, this
           method simply returns :attr:`_element`.

    """

    _element = None

    def _build_element(self):
        return self._element

    element = property(fget=lambda self: self._build_element(),
                       doc="Property that gives the element to be tested.")

    #-----------------------------------------------------------------------

    def run(self, result=None):
        return unittest.TestCase.run(self, result)

    def test_element(self):
        element = self.element
        if not element:
            return

        tester = ElementTester(element)

        try:
            for words, expected_value in self.input_output:
                recognized_value = tester.recognize(words)
                if recognized_value != expected_value:
                    raise TestError(u"Recognition mismatch: input %r gave"
                                    u" recognition %r, instead of expected %r"
                                    % (words, recognized_value,
                                       expected_value))
        except TestError as e:
            self.fail(text_type(e))



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Test suite for Dutch language Integer and Digits classes
============================================================================

"""

from dragonfly.test.infrastructure      import RecognitionFailure
from dragonfly.test.element_testcase    import ElementTestCase
from dragonfly.language.base.integer    import Integer
from dragonfly.language.nl.number       import IntegerContent


#---------------------------------------------------------------------------

class DutchIntegerTestCase(ElementTestCase):
    """ Verify various Dutch integers between 0 and 10**12. """
    def _build_element(self):
        return Integer(content=IntegerContent, min=0, max=10**12 - 1)
    input_output = [
                    ("nul",        0),
                    ("1",          1),
                    ("twee",       2),
                    ("drie",       3),
                    ("vier",       4),
                    ("vijf",       5),
                    ("zes",        6),
                    ("zeven",      7),
                    ("acht",       8),
                    ("negen",      9),
                    ("tien",      10),
                    ("elf",       11),
                    ("twee en twintig",                            22),
                    ("honderd drie en twintig",                   123),
                    ("honderd en vier en twintig",                124),
                    ("zeven honderd negen en tachtig",            789),
                    ("vier en dertig honderd zes en vijftig",    3456),
                   ]



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Characters in text symbols for the English language.
============================================================================

"""

from ...grammar.elements   import (Choice, RuleRef, Repetition, Sequence,
                                   Alternative, Optional, Compound)
from ...grammar.rule_base  import Rule


#---------------------------------------------------------------------------
# Letter names.

letter_names = {
                "Alpha":     "a",
                "Bravo":     "b",
                "Charlie":   "c",
                "Delta":     "d",
                "Echo":      "e",
                "Foxtrot":   "f",
                "Golf":      "g",
                "Hotel":     "h",
                "India":     "i",
                "Juliett":   "j",
                "Kilo":      "k",
                "Lima":      "l",
                "Mike":      "m",
                "November":  "n",
                "Oscar":     "o",
                "Papa":      "p",
                "Quebec":    "q",
                "Romeo":     "r",
                "Sierra":    "s",
                "Tango":     "t",
                "Uniform":   "u",
                "Victor":    "v",
                "Whiskey":   "w",
                "X-ray":     "x",
                "Yankee":    "y",
                "Zulu":      "z",
               }
uppercase_prefix = "(Cap | Shift)"

lowercase_letter_names = letter_names
uppercase_letter_names = dict((name, char.upper())
                              for (name, char) in list(letter_names.items()))


#---------------------------------------------------------------------------
# Digit names.

digit_names = {
               "Zero":    "0",
               "One":     "1",
               "Two":     "2",
               "Three":   "3",
               "Four":    "4",
               "Five":    "5",
               "Six":     "6",
               "Seven":   "7",
               "Eight":   "8",
               "Nine":    "9",
              }


#---------------------------------------------------------------------------
# Symbol names.

symbol_names = {
#                "Enter":                                  "\n",
#                "Tab":                                    "\t",
#                "Space":                                  " ",
                "Exclamation [Mark]":                     "!",
                "At [Sign]":                              "@",
                "(Hash | Pound) [Sign]":                  "#",
                "Dollar [Sign]":                          "$",
                "Percent [Sign]":                         "%",
                "Caret":                                  "^",
                "(Ampersand | and Sign)":                 "&",
                "(Asterisk | Star)":                      "*",
                "(Left | Open) (Paren | Parenthesis)":    "(",
                "(Right | Close) (Paren | Parenthesis)":  ")",
                "(Hyphen | Minus [Sign])":                "-",
                "Underscore":                             "_",
                "(Equal | Equals) [Sign]":                "=",
                "Plus":                                   "+",
                "Backtick":                               "`",
                "Tilde":                                  "~",
                "(Left | Open) Bracket":                  "[",
                "(Right | Close) Bracket":                "]",
                "(Left | Open) Brace":                    "{",
                "(Right | Close) Brace":                  "}",
                "Backslash":                              "\\",
                "[Vertical] Bar":                         "|",
                "Colon":                                  ":",
                "(Apostrophe | Single Quote)":            "'",
                "Double Quote":                           '"',
                "Comma":                                  ",",
                "(Dot | Period | Full Stop)":             ".",
                "Slash":                                  "/",
                "Left Angle Bracket":                     "<",
                "Right Angle Bracket":                    ">",
                "Question [Mark]":                        "?",
               }


#---------------------------------------------------------------------------
# Aggregation of all char names.

char_names = dict(lowercase_letter_names)
char_names.update(digit_names)
char_names.update(symbol_names)


#---------------------------------------------------------------------------
# Helper functions to create special element classes.

class CharChoice(RuleRef):

    def __init__(self, name=None):
        letter_element = Sequence([
                  Optional(Compound(uppercase_prefix, name="upper")),
                  Choice("letter", letter_names),
                ])
        other_choices = {}
        other_choices.update(digit_names)
        other_choices.update(symbol_names)
        other_element = Choice(name="other", choices=other_choices)

        root_element = Alternative([letter_element, other_element])
        root_rule = Rule(element=root_element)
        RuleRef.__init__(self, name=name, rule=root_rule)

    def value(self, node):
        child = node.get_child_by_name("letter")
        if child:
            value = child.value()
            if node.has_child_with_name("upper"):
                value = value.upper()
            return value
        child = node.get_child_by_name("other")
        return child.value()


#---------------------------------------------------------------------------
# Helper functions to create special element classes.

def choice_wrap_class(choices):
    class _ChoiceWrapClass(RuleRef):
        _choices = choices
        def __init__(self, name=None):
            choice_element = Choice(name=None, choices=self._choices)
            choice_rule = Rule(element=choice_element)
            RuleRef.__init__(self, name=name, rule=choice_rule)
    return _ChoiceWrapClass

def choice_series_wrap_class(choices):
    class _ChoiceWrapClass(RuleRef):
        _choices = choices
        def __init__(self, name=None, min=1, max=8):
            choice_element   = Choice(name=None, choices=self._choices)
            choice_rule      = Rule(element=choice_element)
            choice_ref       = RuleRef(rule=choice_rule)
            series_element   = Repetition(child=choice_ref, min=min, max=max)
            series_rule      = Rule(element=series_element)
            RuleRef.__init__(self, name=name, rule=series_rule)
        def value(self, node):
            child_values = RuleRef.value(self, node)
            return "".join(child_values)
    return _ChoiceWrapClass

def element_series_wrap_class(element):
    class _ChoiceWrapClass(RuleRef):
        _element = element
        def __init__(self, name=None, min=1, max=8):
            series_element   = Repetition(child=self._element, min=min, max=max)
            series_rule      = Rule(element=series_element)
            RuleRef.__init__(self, name=name, rule=series_rule)
        def value(self, node):
            child_values = RuleRef.value(self, node)
            return "".join(child_values)
    return _ChoiceWrapClass


#---------------------------------------------------------------------------
# High-level, easy-to-use element classes.

LetterChoice  = choice_wrap_class(lowercase_letter_names)
LetterSeries  = choice_series_wrap_class(lowercase_letter_names)
#CharChoice    = choice_wrap_class(char_names)
#CharSeries    = choice_series_wrap_class(char_names)
CharSeries    = element_series_wrap_class(CharChoice())



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Fundamental element classes
============================================================================

Dragonfly grammars are built up out of a small set of fundamental building
blocks.  These building blocks are implemented by the following *element*
classes:

 - :class:`ElementBase` --
   the base class from which all other element classes are derived
 - :class:`Sequence` --
   sequence of child elements which must all match in the order given
 - :class:`Alternative` --
   list of possibilities of which only one will be matched
 - :class:`Optional` --
   wrapper around a child element which makes the child element optional
 - :class:`Repetition` --
   repetition of a child element
 - :class:`Literal` --
   literal word which must be said exactly by the speaker as given
 - :class:`RuleRef` --
   reference to a :class:`dragonfly.grammar.rule_base.Rule` object;
   this element allows a rule to include (i.e. reference) another rule
 - :class:`ListRef` --
   reference to a :class:`dragonfly.grammar.list.List` object
 - :class:`Impossible` --
   a special element that cannot be recognized
 - :class:`Empty` --
   a special element that is always recognized

The following *element* classes are built up out of the fundamental
classes listed above:

 - :class:`Dictation` --
   free-form dictation; this element matches any words the speaker says,
   and includes facilities for formatting the spoken words with correct
   spacing and capitalization
 - :class:`Modifier` --
   modifies the output of another element by applying a function to it
   following recognition
 - :class:`DictListRef` --
   reference to a :class:`dragonfly.DictList` object; this element is
   similar to the :class:`dragonfly.ListRef` element, except that it returns
   the value associated with the spoken words instead of the spoken words
   themselves
 - :class:`RuleWrap` --
   an element class used to wrap a Dragonfly element into a new private
   rule to be referenced by the same element or other :class:`RuleRef`
   elements

See the following documentation sections for additional information and
usage examples:

 * :ref:`Elements (Object model section) <RefObjectModelElements>`
 * :ref:`RefElementBasicDocTests`

"""

# pylint: disable=abstract-method,no-self-use,too-many-lines
# Suppress a few warnings.

import copy
import itertools
import logging

from six import integer_types, string_types

from .rule_base  import Rule
from .list       import ListBase, DictList

#===========================================================================
# Element base class.

id_generator = itertools.count()

class ElementBase(object):
    """ Base class for all other element classes. """

    name = "uninitialized"

    _log_decode = logging.getLogger("grammar.decode")
    _log_eval = logging.getLogger("grammar.eval")

    def __init__(self, name=None, default=None):
        """
            Constructor argument:
             - *name* (*str*, default: *None*) --
               the name of this element; can be used when interpreting
               complex recognition for retrieving elements by name.
             - *default* (*object*, default: *None*) --
               the default value used if this element is optional and wasn't
               spoken

        """
        if not name:
            name = None
        self.name = name
        self._default = default
        self._id = next(id_generator)

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        if self.name:   name_str = ", name=%r" % self.name
        else:           name_str = ""
        return "%s(...%s)" % (self.__class__.__name__, name_str)

    def _get_children(self):
        """
            Returns an iterable of this element's children.

            This method is used by the :meth:`.children` property, and
            should be overloaded by any derived classes to give
            the correct children element.

            By default, this method returns an empty tuple.

        """
        return ()
    children = property(lambda self: self._get_children(),
                        doc="Iterable of child elements.  (Read-only)")

    def element_tree_string(self):
        """
            Returns a formatted multi-line string representing this
            element and its children.

        """
        indent = "  "
        tree = []
        stack = [(self, 0)]
        while stack:
            element, index = stack.pop()
            if index == 0:
                tree.append((element, len(stack)))
            if len(element.children) > index:
                stack.append((element, index + 1))
                stack.append((element.children[index], 0))
        lines = (indent*depth + str(element) for element, depth in tree)
        return "\n".join(lines)

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def dependencies(self, memo):
        """
            Returns an iterable containing the dependencies of this
            element and of this element's children.

            The dependencies are the objects that are necessary
            for this element.  These include lists and other rules.

        """
        if self._id in memo:
            return []
        memo.add(self._id)
        dependencies = []
        for c in self.children:
            dependencies.extend(c.dependencies(memo))
        return dependencies

    def gstring(self):
        """
            Returns a formatted grammar string of the contents
            of this element and its children.

            The grammar string is of a format similar to that used
            by Natlink to define its grammars.

        """
        raise NotImplementedError("Call to virtual method gstring()"
                                  " in base class ElementBase")

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        """
            Attempt to decode the recognition stored in the given *state*.

        """
        raise NotImplementedError("Call to virtual method decode()"
                                  " in base class ElementBase")

    def value(self, node):
        """
            Determine the semantic value of this element given the
            recognition results stored in the *node*.

            Argument:
             - *node* --
               a :class:`dragonfly.grammar.state.Node` instance
               representing this element within the recognition
               parse tree

            The default behavior of this method is to return
            an iterable containing the recognized words matched
            by this element (i.e. *node.words()*).

        """
        return node.words()

    def has_default(self):
        """
        """
        return self._default is not None

    @property
    def default(self):
        """
        """
        return self._default

    #-----------------------------------------------------------------------
    # Internal utility methods for use by derived classes.

    def _copy_sequence(self, sequence, name, item_types=None):
        """
            Utility function for derived classes that checks that a given
            object is a sequence, copies its contents into a new tuple,
            and checks that each item is of a given type.

        """
        try:
            result = tuple(sequence)
        except TypeError:
            raise TypeError("%s object must be a sequence." % name)

        if not item_types: return result

        invalid = [c for c in result if not isinstance(c, item_types)]
        if invalid:
            raise TypeError("%s object must contain only %s types."
                            "  (Received %s)" % (name, item_types, result))
        return result


#===========================================================================
# Basic structural element classes.

class Sequence(ElementBase):
    """
        Element class representing a sequence of child elements
        which must all match a recognition in the correct order.

        Constructor arguments:
         - *children* (iterable, default: *()*) --
           the child elements of this element
         - *name* (*str*, default: *None*) --
           the name of this element
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

        For a recognition to match, all child elements must match
        the recognition in the order that they were given in the
        *children* constructor argument.

        Example usage::

           >>> from dragonfly.test import ElementTester
           >>> seq = Sequence([Literal("hello"), Literal("world")])
           >>> test_seq = ElementTester(seq)
           >>> test_seq.recognize("hello world")
           ['hello', 'world']
           >>> test_seq.recognize("hello universe")
           RecognitionFailure

    """

    def __init__(self, children=(), name=None, default=None):
        ElementBase.__init__(self, name=name, default=default)
        self._children = self._copy_sequence(children,
                                             "children", ElementBase)

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def _get_children(self):
        """ Returns the child elements contained within the sequence. """
        return self._children

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def dependencies(self, memo):
        if self._id in memo:
            return []
        memo.add(self._id)
        dependencies = []
        for c in self._children:
            dependencies.extend(c.dependencies(memo))
        return dependencies

    def gstring(self):
        return "(" \
             + " ".join([e.gstring() for e in self._children]) \
             + ")"

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        # Special case for an empty sequence.
        if len(self._children) == 0:
            state.decode_success(self)
            yield state
            state.decode_retry(self)
            state.decode_failure(self)
            return

        # Attempt to walk a path through the entire sequence of children
        #  so that each one decodes successfully.
        path = [self._children[0].decode(state)]
        while path:
            # Allow the last child to attempt decoding.
            try: next(path[-1])
            except StopIteration:
                # Last child failed to decode, remove from path to
                #  allowed the one-before-last child to reattempt.
                path.pop()
            else:
                # Last child successfully decoded.
                if len(path) < len(self._children):
                    # Sequence not yet complete, append the next child.
                    path.append(self._children[len(path)].decode(state))
                else:
                    # Sequence complete, all children decoded successfully.
                    state.decode_success(self)
                    yield state
                    state.decode_retry(self)

        # Sequence of children could not all decode successfully: failure.
        state.decode_failure(self)
        return

    def value(self, node):
        """
            The *value* of a :class:`Sequence` is a list containing
            the values of each of its children.

        """
        return [child.value() for child in node.children]


#---------------------------------------------------------------------------

class Optional(ElementBase):
    """
        Element class representing an optional child element.

        Constructor arguments:
         - *child* (*ElementBase*) --
           the child element of this element
         - *name* (*str*, default: *None*) --
           the name of this element
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

        Recognitions always match this element.  If the child element
        does match the recognition, then that result is used.
        Otherwise, this element itself does match but the child
        is not processed.

    """

    def __init__(self, child, name=None, default=None):
        ElementBase.__init__(self, name, default=default)

        if not isinstance(child, ElementBase):
            raise TypeError("Child of %s object must be an"
                            " ElementBase instance." % self)
        self._child = child
        self._greedy = True

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def _get_children(self):
        """ Returns the optional child element. """
        return (self._child, )

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def dependencies(self, memo):
        if self._id in memo:
            return []
        memo.add(self._id)
        return self._child.dependencies(memo)

    def gstring(self):
        return "[" + self._child.gstring() + "]"

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        # pylint: disable=unused-variable
        state.decode_attempt(self)

        # If in greedy mode, allow the child to decode before.
        if self._greedy:
            for result in self._child.decode(state):
                state.decode_success(self)
                yield state
                state.decode_retry(self)

            # Rollback decoding so that the null-decode can be yielded.
            state.decode_rollback(self)

        # Yield the null-decode possibility.
        state.decode_success(self)
        yield state
        state.decode_retry(self)

        # If not in greedy mode, allow the child to decode after.
        if not self._greedy:
            for result in self._child.decode(state):
                state.decode_success(self)
                yield state
                state.decode_retry(self)

        # No more decoding possibilities available, failure.
        state.decode_failure(self)

    def value(self, node):
        """
            The *value* of a :class:`Optional` is the value of its child,
            if the child did match the recognition.  Otherwise the
            *value* is *None*.

        """
        if node.children:
            return node.children[0].value()
        else:
            return None


#---------------------------------------------------------------------------

class Alternative(ElementBase):
    """
        Element class representing several child elements of which only
        one will match.

        Constructor arguments:
         - *children* (iterable, default: *()*) --
           the child elements of this element
         - *name* (*str*, default: *None*) --
           the name of this element
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

        For a recognition to match, at least one of the child elements
        must match the recognition.  The first matching child is
        used.  Child elements are searched in the order they are given
        in the *children* constructor argument.

    """

    def __init__(self, children=(), name=None, default=None):
        ElementBase.__init__(self, name, default=default)
        self._children = self._copy_sequence(children,
                                             "children", ElementBase)

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def _get_children(self):
        """ Returns the alternative child elements. """
        return self._children

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def gstring(self):
        return "(" \
             + " | ".join([e.gstring() for e in self._children]) \
             + ")"

    def dependencies(self, memo):
        if self._id in memo:
            return []
        memo.add(self._id)
        dependencies = []
        for c in self._children:
            dependencies.extend(c.dependencies(memo))
        return dependencies

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        # Special case for an empty list of alternatives.
        if len(self._children) == 0:
            state.decode_success(self)
            yield state
            state.decode_retry(self)
            state.decode_failure(self)
            return

        # Iterate through the children.
        for child in self._children:

            # Iterate through this child's possible decoding states.
            # pylint: disable=unused-variable
            for result in child.decode(state):
                state.decode_success(self)
                yield state
                state.decode_retry(self)

            # Rollback to the alternative's original state, so that the
            #  next child starts decoding without interference from the
            #  previous child.
            state.decode_rollback(self)

        # None of the children could decode successfully: failure.
        state.decode_failure(self)
        return

    def value(self, node):
        """
            The *value* of an :class:`Alternative` is the value of its
            child that matched the recognition.

        """
        return node.children[0].value()


#---------------------------------------------------------------------------

class Repetition(Sequence):
    """
        Element class representing a repetition of one child element.

        Constructor arguments:
         - *child* (*ElementBase*) --
           the child element of this element
         - *min* (*int*, default: *1*) --
           the minimum number of times that the child element must
           be recognized; may be 0
         - *max* (*int*, default: *None*) --
           the maximum number of times that the child element must
           be recognized; if *None*, the child element must be recognized
           exactly *min* times (i.e. *max = min + 1*)
         - *name* (*str*, default: *None*) --
           the name of this element
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken
         - *optimize* (*bool*, default: *True*) --
           whether the engine's compiler should compile the element
           optimally

        For a recognition to match, at least one of the child elements
        must match the recognition.  The first matching child is
        used.  Child elements are searched in the order they are given
        in the *children* constructor argument.

        If the *optimize* argument is set to *True*, the compiler will
        ignore the *min* and *max* limits to reduce grammar complexity. If
        the number of repetitions recognized is more than the *max* value,
        the rule will fail to match.

    """

    # pylint: disable=redefined-builtin,unused-variable

    def __init__(self, child, min=1, max=None, name=None, default=None,
                 optimize=True):
        if not isinstance(child, ElementBase):
            raise TypeError("Child of %s object must be an"
                            " ElementBase instance." % self)
        assert isinstance(min, integer_types)
        assert max is None or isinstance(max, integer_types)
        assert max is None or min < max, "min must be less than max"

        self._child = child
        self._min = min
        if max is None: self._max = min + 1
        else:           self._max = max
        self._optimize = optimize

        optional_length = self._max - self._min - 1
        if optional_length > 0:
            element = Optional(child)
            for index in range(optional_length-1):
                element = Optional(Sequence([child, element]))

            if self._min >= 1:
                children = [child] * self._min + [element]
            else:
                children = [element]
        elif self._min > 0:
            children = [child] * self._min
        else:
            raise ValueError("Repetition not allowed to be empty.")

        Sequence.__init__(self, children, name=name, default=default)

    min = property(
        lambda self: self._min,
        doc="The minimum number of times that the child element must be "
        "recognized; may be 0. (Read-only)"
    )

    max = property(
        lambda self: self._max,
        doc="The maximum number of times that the child element must be "
        "recognized; if *None*, the child element must be "
        "recognized exactly *min* times (i.e. *max = min + 1*). "
        "(Read-only)"
    )

    optimize = property(
        lambda self: self._optimize,
        doc="Whether the engine's compiler should compile the element "
        "optimally. (Read-only)"
    )

    def dependencies(self, memo):
        if self._id in memo:
            return []
        memo.add(self._id)
        return self._child.dependencies(memo)

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def get_repetitions(self, node):
        """
            Returns a list containing the nodes associated with
            each repetition of this element's child element.

            Argument:
             - *node* (*Node*) --
               the parse tree node associated with this repetition element;
               necessary for searching for child elements within the parse
               tree

        """
        repetitions = []
        for index in range(self._min):
            element = node.children[index]
            if element.actor != self._child:
                raise TypeError("Invalid child of %s: %s" \
                    % (self, element.actor))
            repetitions.append(element)

        if self._max - self._min > 1:
            optional = node.children[-1]
            while optional.children:
                child = optional.children[0]
                if isinstance(child.actor, Sequence):
                    assert len(child.children) == 2
                    element, optional = child.children
                    if element.actor != self._child:
                        raise TypeError("Invalid child of %s: %s" \
                            % (self, element.actor))
                    repetitions.append(element)
                elif child.actor == self._child:
                    repetitions.append(child)
                    break
                else:
                    raise TypeError("Invalid child of %s: %s" \
                        % (self, child.actor))

        return repetitions

    def value(self, node):
        """
            The *value* of a :class:`Repetition` is a list containing
            the values of its child.

            The length of this list is equal to the number of times
            that the child element was recognized.

        """
        repetitions = self.get_repetitions(node)
        return [r.value() for r in repetitions]


#---------------------------------------------------------------------------

class Literal(ElementBase):
    """
        Element class representing one or more literal words which must be
        said exactly by the speaker as given.

        Quoted words can be used to potentially improve accuracy. This
        currently only has an effect if using the Natlink SR engine
        back-end.

        Constructor arguments:
         - *text* (*str*) --
           the words to be said by the speaker
         - *name* (*str*, default: *None*) --
           the name of this element
         - *value* (*object*, default: *the recognized words*) --
           value returned when this element is successfully decoded
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken
         - *quote_start_str* (*str*, default: \") --
           the string used for specifying the start of quoted words.
         - *quote_end_str* (*str*, default: \") --
           the string used for specifying the end of quoted words.
         - *strip_quote_strs* (*bool*, default: *True*) --
           whether the start and end quote strings should be stripped from
           this element's word lists.

    """

    def __init__(self, text, name=None, value=None, default=None,
                 quote_start_str='"', quote_end_str='"',
                 strip_quote_strs=True):
        ElementBase.__init__(self, name, default=default)
        self._value = value

        if not isinstance(text, string_types):
            raise TypeError("Text of %s object must be a"
                            " string." % self)

        # Construct the words and extended words lists. The latter includes
        # quoted words as single items.
        words = []
        words_ext = []
        current_quoted_sequence = []
        for word in text.split():
            # Begin quoted words.
            if word.startswith(quote_start_str):
                current_quoted_sequence.append(word)

            # End quoted words.
            elif current_quoted_sequence and word.endswith(quote_end_str):
                current_quoted_sequence.append(word)
                quoted_words = " ".join(current_quoted_sequence)

                # Strip quote start and end strings if specified.
                if strip_quote_strs:
                    quoted_words = quoted_words[len(quote_start_str):
                                                len(quote_end_str) * -1]

                # Add the words to both lists.
                words.extend(quoted_words.split())
                words_ext.append(quoted_words)

                # Clear current sequence list.
                del current_quoted_sequence[:]

            # Continuing quoted words.
            elif current_quoted_sequence:
                current_quoted_sequence.append(word)

            # Unquoted words.
            else:
                words.append(word)
                words_ext.append(word)

        # Handle unfinished quoted words sequence by treating it as normal.
        if current_quoted_sequence:
            words.extend(current_quoted_sequence)
            words_ext.extend(current_quoted_sequence)

        # Set both lists.
        self._words = words
        self._words_ext = words_ext

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.words)

    words = property(
        lambda self: self._words,
        doc="The words to be said by the speaker."
    )

    words_ext = property(
        lambda self: self._words_ext,
        doc="The words to be said by the speaker, with any quoted words as "
        "single items. This is extends the :py:attr:`~words` property."
    )

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def dependencies(self, memo):
        return []

    def gstring(self):
        return " ".join(self._words)

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        # Iterate through this element's words.
        # If all match, success.  Else, failure.
        if state.engine and state.engine.quoted_words_support:
            words = self._words_ext
        else:
            words = self._words
        for i in range(len(words)):
            word = state.word(i)

            # If word isn't None, make it lowercase.
            if word:
                word = word.lower()

            # If the words are not the same, then this is a decode failure.
            if word != words[i].lower():
                state.decode_failure(self)
                return

        # All words matched, success.
        state.next(len(words))
        state.decode_success(self)
        yield state
        state.decode_retry(self)

        # Only one decoding possibility, failure on retry.
        state.decode_failure(self)
        return

    def value(self, node):
        if self._value is None:
            return u" ".join(node.words())
        else:
            return self._value

#---------------------------------------------------------------------------

class RuleRef(ElementBase):
    """
        Element class representing a reference to another Dragonfly rule.

        Constructor arguments:
         - *rule* (*Rule*) --
           the Dragonfly rule to reference
         - *name* (*str*, default: *None*) --
           the name of this element
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

    """

    def __init__(self, rule, name=None, default=None):
        ElementBase.__init__(self, name, default=default)

        if not isinstance(rule, Rule):
            raise TypeError("Rule object of %s object must be a"
                            " Dragonfly rule." % self)
        self._rule = rule

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        if not hasattr(self, "_rule"):
            return ElementBase.__repr__(self)
        return '%s(%s)' % (self.__class__.__name__, self._rule.name)

    rule = property(lambda self: self._rule)

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def dependencies(self, memo):
        if self._id in memo:
            return []
        memo.add(self._id)
        return [self._rule] + self._rule.dependencies(memo)

    def gstring(self):
        return "<" + self._rule.name + ">"


    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        # Allow the rule to attempt decoding.
        # pylint: disable=unused-variable
        for result in self._rule.decode(state):
            state.decode_success(self)
            yield state
            state.decode_retry(self)

        # The rule failed to deliver a valid decoding, failure.
        state.decode_failure(self)

    def value(self, node):
        return node.children[0].value()


#---------------------------------------------------------------------------

class ListRef(ElementBase):
    """
        Element class representing a reference to a Dragonfly List.

        Constructor arguments:
         - *name* (*str*, default: *None*) --
           the name of this element
         - *list* (*ListBase*) --
           the Dragonfly List to reference
         - *key* (*object*, default: *None*) --
           key to differentiate between list references at runtime
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

    """

    # pylint: disable=redefined-builtin
    def __init__(self, name, list, key=None, default=None):
        self._list = None
        self._key = None

        ElementBase.__init__(self, name=name, default=default)

        if not isinstance(list, ListBase):
            raise TypeError("List argument to %s constructor must be a"
                            " Dragonfly list." % self.__class__.__name__)
        self._list = list
        self._key = key

    #-----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        arguments = []
        if self._list is not None:
            arguments.append(repr(self._list.name))
        if self._key:
            arguments.append("key=%r" % self._key)
        return "%s(%s)" % (self.__class__.__name__, ", ".join(arguments))

    list = property(lambda self: self._list)

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def dependencies(self, memo):
        if self._id in memo:
            return []
        memo.add(self._id)
        return [self._list]

    def gstring(self):
        return "{" + self._list.name + "}"

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        # If the next word(s) is/are in the list, success.
        delta = 0
        word = state.word()
        while 1:
            if word in self._list:
                state.next(delta + 1)
                state.decode_success(self)
                yield state
                state.decode_retry(self)
            delta += 1
            next = state.word(delta)
            if next is None:
                break
            word += " " + next

        # If the word is not in the list, or on retry, failure.
        state.decode_failure(self)

    def value(self, node):
        words = node.words()
        return " ".join(words)

#---------------------------------------------------------------------------

class DictListRef(ListRef):
    """
        Element class representing a reference to a Dragonfly DictList.

        Constructor arguments:
         - *name* (*str*, default: *None*) --
           the name of this element
         - *dict* (*DictList*) --
           the Dragonfly DictList to reference
         - *key* (*object*, default: *None*) --
           key to differentiate between list references at runtime
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

    """

    # pylint: disable=redefined-builtin
    def __init__(self, name, dict, key=None, default=None):
        if not isinstance(dict, DictList):
            raise TypeError("Dict object of %s object must be a Dragonfly "
                            "DictList." % self.__class__.__name__)
        ListRef.__init__(self, name, dict, key, default=default)

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def value(self, node):
        key = ListRef.value(self, node)
        return self._list[key]

#---------------------------------------------------------------------------

class Empty(ElementBase):
    """
        Element class representing something that is always recognized.

        Constructor arguments:
         - *name* (*str*, default: *None*) --
           the name of this element
         - *value* (*object*, default: *True*) --
           value returned when this element is successfully decoded (always)

        Empty elements are equivalent to children of :class:`Optional`
        elements.

    """

    def __init__(self, name=None, value=True, default=None):
        self._value = value
        ElementBase.__init__(self, name, default=default)

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def gstring(self):
        return "<Empty()>"

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        state.decode_success(self)
        yield state
        state.decode_retry(self)

        state.decode_failure(self)

    def value(self, node):
        return self._value


#===========================================================================
# Slightly more complex element classes.

class Dictation(ElementBase):
    """
        Element class representing free dictation.

        Constructor arguments:
         - *name* (*str*, default: *None*) --
           the name of this element
         - *format* (*bool*, default: *True*) --
           whether the value returned should be a
           :class:`DictationContainerBase` object. If *False*, then the
           returned value is a list of the recognized words
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

        Returns a string-like :class:`DictationContainerBase` object
        containing the recognised words.
        By default this is formatted as a lowercase sentence.
        Alternative formatting can be applied by calling string methods like
        `replace` or `upper` on a :class:`Dictation` object, or by passing
        an arbitrary formatting function (taking and returning a string) to
        the `apply` method.

        Camel case text can be produced using the `camel` method. For
        example:

        .. code:: python

            str_func = lambda s: s.upper()
            Dictation("formattedtext").apply(str_func)
            Dictation("snake_text").lower().replace(" ", "_")
            Dictation("camelText").camel()
    """

    # pylint: disable=redefined-builtin
    def __init__(self, name=None, format=True, default=None):
        ElementBase.__init__(self, name, default=default)
        self._format_words = format
        self._string_methods = []

    def __repr__(self):
        if self.name:
            return "%s(%r)" % (self.__class__.__name__, self.name)
        else:
            return "%s()" % (self.__class__.__name__)

    def __getattr__(self, name):
        if isinstance(name, str) and name[:2] == name[-2:] == '__':
            # skip non-existing dunder method lookups
            raise AttributeError(name)
        def call(*args, **kwargs):
            self._string_methods.append((name, args, kwargs))
            return self
        return call

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def gstring(self):
        return "<Dictation()>"

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        # Check that at least one word has been dictated, otherwise fail.
        if state.rule() != "dgndictation":
            state.decode_failure(self)
            return

        # Determine how many words have been dictated.
        count = 1
        while state.rule(count) == "dgndictation":
            count += 1

        # Yield possible states where the number of dictated words
        # gobbled is decreased by 1 between yields.
        for i in range(count, 0, -1):
            state.next(i)
            state.decode_success(self)
            yield state
            state.decode_retry(self)
            state.decode_rollback(self)

        # None of the possible states were accepted, failure.
        state.decode_failure(self)
        return

    def value(self, node):
        if self._format_words:
            return node.engine.DictationContainer(node.words(),
                                                  self._string_methods)
        else:
            return node.words()

#---------------------------------------------------------------------------

class Modifier(Alternative):
    """
        Element allowing direct modification of the output of another
        element at recognition time.

        Constructor arguments:
            - *element* (*Element*) -- The element to be recognised, e.g.
              :class:`Dictation` or :class:`Repetition`, with appropriate
              arguments passed.
            - *modifier* (*function*) -- A function to be applied to the
              value of this element when it is recognised.

        Examples:

        .. code:: python

            # Recognises an integer, returns the integer plus one
            Modifier(IntegerRef("plus1", 1, 20), lambda n: n+1)

            # Recognises a series of integers, returns them separated by
            # commas as a string
            int_rep = Repetition(IntegerRef("", 0, 10), min=1, max=5,
                                 name="num_seq")
            Modifier(int_rep, lambda r: ", ".join(map(str, r)))

    """
    def __init__(self, element, modifier=None):
        self._modifier = modifier
        Alternative.__init__(self, children=(element,), name=element.name,
                             default=element.default)

    def value(self, node):
        initial_value = Alternative.value(self, node)
        if self._modifier:
            return self._modifier(initial_value)
        else:
            return initial_value

#---------------------------------------------------------------------------

class Impossible(ElementBase):
    """
        Element class representing speech that cannot be recognized.

        Constructor arguments:
         - *name* (*str*, default: *None*) --
           the name of this element

        Using an :class:`Impossible` element in a Dragonfly rule makes it
        impossible to recognize via speech or mimicry.

    """

    def __init__(self, name=None):
        ElementBase.__init__(self, name)

    #-----------------------------------------------------------------------
    # Methods for load-time setup.

    def gstring(self):
        return "<Impossible()>"

    #-----------------------------------------------------------------------
    # Methods for runtime recognition processing.

    def decode(self, state):
        state.decode_attempt(self)

        # Impossible elements always fail to decode.
        state.decode_failure(self)
        return

        # Turn this method into a generator by using 'yield'. This works
        # even though the statement is unreachable.
        # pylint: disable=unreachable
        yield state

    def value(self, node):
        return self


#---------------------------------------------------------------------------

class RuleWrap(RuleRef):
    """
        Element class to wrap a Dragonfly element into a new private rule
        to be referenced by this element or others.

        :class:`RuleWrap` is a sub-class of :class:`RuleRef`, so
        :class:`RuleWrap` elements can be used in the same way as
        :class:`RuleRef` elements.

        Constructor arguments:
         - *name* (*str*, default: *None*) --
           the name of this element
         - *element* (*Element*) --
           the Dragonfly element to be wrapped
         - *default* (*object*, default: *None*) --
           the default value used if this element is optional and wasn't
           spoken

        Examples:

        .. code:: python

           # For saying and processing a Choice element two times.
           letter = RuleWrap("letter1", Choice("", {
               "alpha": "a",
               "bravo": "b",
               "charlie": "c"
           }))
           letter_extras = [
               letter,
               RuleRef(letter.rule, "letter2"),
               RuleRef(letter.rule, "letter3")
           ]

    """

    _next_id = 0

    def __init__(self, name, element, default=None):
        rule_name = "_%s_%02d" % (self.__class__.__name__, RuleWrap._next_id)
        RuleWrap._next_id += 1
        rule = Rule(name=rule_name, element=element, exported=False)
        RuleRef.__init__(self, rule=rule, name=name, default=default)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Command module loading classes
============================================================================

"""

import os.path
import logging

import six

# --------------------------------------------------------------------------
# Command module class; wraps a single command-module.


class CommandModule(object):

    _log = logging.getLogger("module")

    def __init__(self, path):
        self._path = os.path.abspath(path)
        self._namespace = None
        self._loaded = False

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__,
                           os.path.basename(self._path))

    @property
    def loaded(self):
        return self._loaded

    def load(self):
        self._log.info("%s: Loading module: '%s'", self, self._path)

        # Prepare namespace in which to execute the
        namespace = {"__file__": self._path}

        # Attempt to execute the module; handle any exceptions.
        try:
            # pylint: disable=exec-used
            # Read from the file in binary mode to avoid decoding issues.
            with open(self._path, "rb") as f:
                contents = f.read()
            exec(compile(contents, self._path, 'exec'), namespace)
        except Exception as e:
            self._log.exception("%s: Error loading module: %s", self, e)
            self._loaded = False
            return

        self._loaded = True
        self._namespace = namespace

    def unload(self):
        self._log.info("%s: Unloading module: '%s'", self, self._path)

    def check_freshness(self):
        pass


# --------------------------------------------------------------------------
# Command module directory class.

class CommandModuleDirectory(object):

    _log = logging.getLogger("directory")

    def __init__(self, path, excludes=None, recursive=False):
        if excludes is None:
            excludes = []

        self._path = os.path.abspath(path)
        self._excludes = excludes
        self._recursive = recursive
        self._modules = {}

    def load(self):
        valid_paths = self._get_valid_paths()

        # Remove any deleted modules.
        for path, module_ in self._modules.items():
            if path not in valid_paths:
                del self._modules[path]
                module_.unload()

        # Add any new modules.
        for path in valid_paths:
            if path not in self._modules:
                if os.path.isfile(path):
                    module_ = CommandModule(path)
                elif os.path.isdir(path):
                    module_ = CommandModuleDirectory(path, self._excludes,
                                                     self._recursive)
                module_.load()
                self._modules[path] = module_
            else:
                module_ = self._modules[path]
                module_.check_freshness()

    @property
    def loaded(self):
        return not any([
            module_ for module_ in self._modules.values()
            if not module_.loaded
        ])

    def _get_valid_paths(self):
        self._log.info("Looking for command modules here: %s", self._path)
        valid_paths = []
        for filename in os.listdir(self._path):
            path = os.path.abspath(os.path.join(self._path, filename))
            if not (self._recursive or os.path.isfile(path)):
                continue
            if path in self._excludes:
                continue

            # Only apply _*.py to files, not directories.
            is_file = os.path.isfile(path)
            if is_file and not (os.path.basename(path).startswith("_") and
                                os.path.splitext(path)[1] == ".py"):
                continue
            valid_paths.append(path)
        self._log.info("Valid paths: %s", ", ".join(valid_paths))
        return valid_paths

    def unload(self):
        self._log.info("%s: Unloading directory: '%s'", self, self._path)
        for path, module_ in tuple(self._modules.items()):
            del self._modules[path]
            module_.unload()

    def check_freshness(self):
        pass



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file contains the base interface to the system clipboard.
"""

# pylint: disable=W0622
# Suppress warnings about redefining the built-in 'format' function.

import contextlib
import functools
import locale
import logging
import os
import re
import time

from six import text_type, binary_type, integer_types


#===========================================================================


@functools.total_ordering
class BaseClipboard(object):
    """
    Base clipboard class.
    """

    _log = logging.getLogger("clipboard")

    format_text      = 1
    format_unicode   = 13
    format_hdrop     = 15
    format_names = {
        format_text:     "text",
        format_unicode:  "unicode",
        format_hdrop:    "hdrop",
    }

    #-----------------------------------------------------------------------

    @classmethod
    def get_system_text(cls):
        """
        Retrieve the system clipboard text.
        """
        raise NotImplementedError()

    @classmethod
    def set_system_text(cls, content):
        """
            Set the system clipboard text.

            Arguments:
             - *content* (string) -- the clipboard contents to set.

            If *None* is given as the *content*, text on the system
            clipboard will be cleared.

        """
        raise NotImplementedError()

    @classmethod
    def clear_clipboard(cls):
        """
        Clear the system clipboard.
        """
        raise NotImplementedError()

    @classmethod
    def _clipboard_formats_changed(cls, formats, clipboard1, clipboard2):
        # Check if the content of any specified format has changed.
        # Use all formats, if specified.
        if formats == "all":
            return clipboard1 != clipboard2
        result = False
        for format in formats:
            format_available = clipboard1.has_format(format)
            result = (
                format_available != clipboard2.has_format(format) or
                format_available and clipboard1.get_format(format)
                != clipboard2.get_format(format))
            if result:
                break
        return result

    @classmethod
    def wait_for_change(cls, timeout, step=0.001, formats=None,
                        initial_clipboard=None):
        """
            Wait (poll) for the system clipboard to change.

            This is a blocking method which returns whether or not the
            system clipboard changed within a specified timeout period.

            Arguments:
             - *timeout* (float) -- timeout in seconds.
             - *step* (float, default: 0.001) -- number of seconds between
               each check.
             - *formats* (iterable, default: None) -- if not None, only
               changes to the given content formats will register.  If None,
               all formats will be observed.
             - *initial_clipboard* (Clipboard, default: None) -- if a
               clipboard is given, the method will wait until the system
               clipboard differs from the instance's contents.

        """
        # By default, this method retrieves the system clipboard every
        #  *step* seconds until the contents change.  This method should be
        #  overridden by the sub-class if there is a more efficient way for
        #  the platform.
        if not initial_clipboard:
            initial_clipboard = cls(from_system=True)
        clipboard2 = cls()
        timeout = time.time() + float(timeout)
        step = float(step)
        if isinstance(formats, integer_types):
            formats = (formats,)
        elif formats:
            for format in formats:
                if not isinstance(format, integer_types):
                    raise TypeError("Invalid clipboard format: %r"
                                    % format)
        result = False
        while time.time() < timeout:
            # Check if the content of any relevant format has changed.
            clipboard2.copy_from_system()
            formats_to_compare = formats if formats else "all"
            result = cls._clipboard_formats_changed(formats_to_compare,
                                                    initial_clipboard,
                                                    clipboard2)

            if result:
                break

            # Failure. Try again after *step* seconds.
            time.sleep(step)
        return result

    @classmethod
    @contextlib.contextmanager
    def synchronized_changes(cls, timeout, step=0.001, formats=None,
                             initial_clipboard=None):
        """
            Context manager for synchronizing local and system clipboard
            changes.  This takes the same arguments as the
            :meth:`wait_for_change` method.

            Arguments:
             - *timeout* (float) -- timeout in seconds.
             - *step* (float, default: 0.001) -- number of seconds between
               each check.
             - *formats* (iterable, default: None) -- if not None, only
               changes to the given content formats will register.  If None,
               all formats will be observed.
             - *initial_clipboard* (Clipboard, default: None) -- if a
               clipboard is given, the method will wait until the system
               clipboard differs from the instance's contents.

            Use with a Python 'with' block::

               from dragonfly import Clipboard, Key

               # Copy the selected text with Ctrl+C and wait until a system
               #  clipboard change is detected.
               timeout = 3
               with Clipboard.synchronized_changes(timeout):
                   Key("c-c", use_hardware=True).execute()

               # Retrieve the system text.
               text = Clipboard.get_system_text()

        """
        # Save the current clipboard contents, if necessary.
        if initial_clipboard:
            initial_clipboard = cls(from_system=True)
        try:
            # Yield for clipboard operations.
            yield
        finally:
            # Wait for the system clipboard to change.
            cls.wait_for_change(timeout, step, formats, initial_clipboard)

    #-----------------------------------------------------------------------

    @classmethod
    def _convert_format_text(cls, content):
        format_name = cls.format_names[cls.format_text]
        if not isinstance(content, (text_type, binary_type)):
            raise TypeError("Invalid content for format %s: (%r)"
                            % (format_name, content,))

        # Use a binary string for CF_TEXT content.
        if isinstance(content, text_type):
            encoding = locale.getpreferredencoding()
            content = content.encode(encoding)

        return content

    @classmethod
    def _convert_format_unicode(cls, content):
        format_name = cls.format_names[cls.format_unicode]
        if not isinstance(content, (text_type, binary_type)):
            raise TypeError("Invalid content for format %s: (%r)"
                            % (format_name, content,))

        # Use a text string for CF_UNICODETEXT content.
        if isinstance(content, binary_type):
            encoding = locale.getpreferredencoding()
            content = content.decode(encoding)

        return content

    @classmethod
    def _convert_format_hdrop(cls, content):
        # The CF_HDROP clipboard format is for copied file paths.
        format_name = cls.format_names[cls.format_hdrop]
        result = content

        # Convert string content into a list of file paths.  String content
        #  must be a list of file paths separated by new lines and/or null
        #  characters.  The following example string is acceptable:
        #  "c:\\temp1.txt\0c:\\temp2.txt\0\0".
        if isinstance(content, (text_type, binary_type)):
            delimiter = u"[\0\r\n]"
            if isinstance(content, binary_type):
                delimiter = delimiter.encode()

            # Convert string content into a list and remove empty
            #  strings.
            result = re.split(delimiter, content)
            result = [string for string in result if string]

        # Verify that the list/tuple items are non-empty strings.
        elif isinstance(content, (list, tuple)):
            string_items = all([
                isinstance(item, (text_type, binary_type)) and item
                for item in content
            ])
            if not string_items:
                raise TypeError("Invalid content type for format %s: (%r)"
                                % (format_name, content))

        # Only strings, tuples and lists are accepted for CF_HDROP
        #  content.
        else:
            raise TypeError("Invalid content type for format %s: (%r)"
                            % (format_name, content))

        # Remove the file scheme from strings if it is present.
        file_paths = []
        for string in result:
            file_scheme = (u"file://" if isinstance(string, text_type)
                           else b"file://")
            if string.startswith(file_scheme):
                string = string[7:]
            file_paths.append(string)
        result = file_paths

        # Empty lists/tuples are not acceptable.
        if not result:
            raise ValueError("Invalid content value for format %s: (%r)"
                             % (format_name, content))

        # Verify that strings in the content list are existing, absolute
        #  file paths.  Relative paths are not accepted because they do not
        #  make sense without the context of a working directory. The system
        #  clipboard does not typically have such a context.
        file_paths_exist = all([
            os.path.isabs(item) and os.path.exists(item)
            for item in result
        ])
        if not (file_paths_exist and result):
            raise ValueError("Invalid content value for format %s: (%r)"
                             % (format_name, content))

        # Use a tuple for CF_HDROP content.
        if isinstance(result, list):
            result = tuple(result)

        return result

    @classmethod
    def convert_format_content(cls, format, content):
        """
            Convert content for the given *format*, if necessary, and return
            it.

            This method operates on the following formats:
             - *text* -- encodes content to a binary string, if necessary
               and if possible.
             - *unicode* -- decodes content to a text string, if necessary
               and if possible.
             - *hdrop* -- converts content into a tuple of file paths, if
               necessary and if possible.

               String content must be a list of existing, absolute file
               paths separated by new lines and/or null characters.  All
               specified file paths must be absolute paths referring to
               existing files on the system.

            If the content cannot be converted for the given *format*, an
            error is raised.

            Arguments:
             - *format* (int) -- the clipboard format to convert.
             - *content* (string) -- the clipboard contents to convert.

        """
        convert_method = {
            cls.format_text:     cls._convert_format_text,
            cls.format_unicode:  cls._convert_format_unicode,
            cls.format_hdrop:    cls._convert_format_hdrop,
        }.get(format)
        if convert_method is not None:
            content = convert_method(content)

        # Return the content.
        return content

    #-----------------------------------------------------------------------

    def __init__(self, contents=None, text=None, from_system=False):
        self._contents = {}

        # If requested, retrieve current system clipboard contents.
        if from_system:
            self.copy_from_system()

        # Process given contents for this Clipboard instance.
        if contents:
            try:
                contents = dict(contents)

                # Enumerate and add contents, converting contents if
                #  necessary.
                for format, content in contents.items():
                    content = self.convert_format_content(format, content)
                    self._contents[format] = content
            except Exception as e:
                raise TypeError("Invalid contents: %s (%r)" % (e, contents))

        # Handle special case of text content.
        if text is not None:
            text = self.convert_format_content(self.format_unicode, text)
            self._contents[self.format_unicode] = text

    def __eq__(self, other):
        formats = self.get_available_formats()
        if formats != other.get_available_formats():
            return False
        for format in formats:
            if self.get_format(format) != other.get_format(format):
                return False
        return True

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        return not self == other

    def __repr__(self):
        arguments = []
        skip = []
        if self.format_unicode in self._contents:
            arguments.append("unicode=%r"
                             % self._contents[self.format_unicode])
            skip.append(self.format_unicode)
        elif self.format_text in self._contents:
            arguments.append("text=%r" % self._contents[self.format_text])
            skip.append(self.format_text)
        for format in sorted(self._contents.keys()):
            if format in skip:
                continue
            if format in self.format_names:
                arguments.append(self.format_names[format])
            else:
                arguments.append(repr(format))
        arguments = ", ".join(str(a) for a in arguments)
        return "%s(%s)" % (self.__class__.__name__, arguments)

    def copy_from_system(self, formats=None, clear=False):
        """
            Copy the system clipboard contents into this instance.

            Arguments:
             - *formats* (iterable, default: None) -- if not None, only the
               given content formats will be retrieved.  If None, all
               available formats will be retrieved.
             - *clear* (boolean, default: False) -- if true, the system
               clipboard will be cleared after its contents have been
               retrieved.

        """
        raise NotImplementedError()

    def copy_to_system(self, clear=True):
        """
            Copy the contents of this instance into the system clipboard.

            Arguments:
             - *clear* (boolean, default: True) -- if true, the system
               clipboard will be cleared before this instance's contents are
               transferred.

        """
        raise NotImplementedError()

    def get_available_formats(self):
        """
            Retrieve a list of this instance's available formats.

            The preferred text format, if available, will always be the
            first on the list followed by any remaining formats in
            numerical order.

        """
        # Return a list of available formats using the same order as
        #  __repr__().
        formats = []
        if self.format_unicode in self._contents:
            formats.append(self.format_unicode)
        elif self.format_text in self._contents:
            formats.append(self.format_text)
        for format in sorted(self._contents.keys()):
            if format in formats:
                continue
            formats.append(format)

        return formats

    def has_format(self, format):
        """
            Determine whether this instance has content for the given
            *format*.

            Arguments:
             - *format* (int) -- the clipboard format to look for.

        """
        return format in self._contents

    def get_format(self, format):
        """
            Retrieved this instance's content for the given *format*.

            Arguments:
             - *format* (int) -- the clipboard format to retrieve.

            If the given *format* is not available, a *ValueError*
            is raised.

        """
        try:
            return self._contents[format]
        except KeyError:
            raise ValueError("Clipboard format not available: %r"
                             % format)

    def set_format(self, format, content):
        """
            Set this instance's content for the given *format*.

            Arguments:
             - *format* (int) -- the clipboard format to set.
             - *content* (string) -- the clipboard contents to set.

            If the given *format* is not available, a *ValueError*
            is raised.

            If *None* is given as the *content*, any content stored
            for the given *format* will be cleared.
        """
        if content is None:
            self._contents.pop(format, None)
        else:
            content = self.convert_format_content(format, content)
            self._contents[format] = content

    def has_text(self):
        """ Determine whether this instance has text content. """
        return (self.format_unicode in self._contents
                or self.format_text in self._contents)

    def get_text(self):
        """
            Retrieve this instance's text content.  If no text content
            is available, this method returns *None*.

        """
        if self.format_unicode in self._contents:
            return self._contents[self.format_unicode]
        elif self.format_text in self._contents:
            return self._contents[self.format_text]
        else:
            return None

    def set_text(self, content):
        """
            Set the text content for this instance.

            Arguments:
             - *content* (string) -- the text content to set.

            If *None* is given as the *content*, any text content
            stored in this instance will be cleared.

        """
        # Clear text content for this instance, if requested.
        if content is None:
            self._contents.pop(self.format_text, None)
            self._contents.pop(self.format_unicode, None)
            return

        if isinstance(content, binary_type):
            format = self.format_text
        else:
            format = self.format_unicode

        # Set this instance's content for the appropriate format.
        self.set_format(format, content)

    text = property(get_text, set_text)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Rule class
============================================================================

"""

# pylint: disable=unused-variable,too-many-instance-attributes,no-self-use

import logging

from .context import Context
from ..error import GrammarError


class Rule(object):
    """
        Rule class for implementing complete or partial voice-commands.

        This rule class represents a voice-command or part of a voice-
        command.  It contains a root element, which defines the language
        construct of this rule.

        Constructor arguments:
         - *name* (*str*) -- name of this rule.  If *None*, a unique
           name will automatically be generated.
         - *element* (*Element*) --
           root element for this rule
         - *context* (*Context*, default: *None*) --
           context within which to be active.  If *None*, the rule will
           always be active when its grammar is active.
         - *imported* (boolean, default: *False*) --
           if true, this rule is imported from outside its grammar
         - *exported* (boolean, default: *True*) --
           if true, this rule is a complete top-level rule which can be
           spoken by the user.  This should be *True* for voice-commands
           that the user can speak.

        The *self._log* logger objects should be used in methods of
        derived classes for logging purposes.  It is a standard logger
        object from the *logger* module in the Python standard library.

    """

    _log_load   = logging.getLogger("grammar.load")
    _log_eval   = logging.getLogger("grammar.eval")
    _log_proc   = logging.getLogger("grammar.process")
    _log        = logging.getLogger("rule")
    _log_begin  = logging.getLogger("rule")

    # Counter ID used for anonymous rules to give them a unique name.
    _next_anonymous_id = 0

    def __init__(self, name=None, element=None, context=None,
                 imported=False, exported=True):
        # The default argument for *element* is NOT acceptable; this
        #  construction is used for backwards compatibility and argument
        #  order.
        self._element = element
        self._imported = imported
        self._exported = exported

        # Generate a unique name for this rule if none is given.
        if not name:
            name = "_anonrule_%03d_%s" % (Rule._next_anonymous_id,
                                          self.__class__.__name__)
            Rule._next_anonymous_id += 1
        self._name = name

        self._active = None
        self._enabled = True
        if not (isinstance(context, Context) or context is None):
            raise TypeError("context must be either a Context object or "
                            "None")
        self._context = context
        self._grammar = None

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self._name)

    #-----------------------------------------------------------------------
    # Protected attribute access.

    name     = property(lambda self: self._name,
                        doc="This rule's name.  (Read-only)")
    element  = property(lambda self: self._element,
                        doc="This rule's root element.  (Read-only)")
    exported = property(lambda self: self._exported,
                        doc="This rule's exported status. See"
                            " :ref:`RefObjectModelRulesExported` for"
                            " more info.  (Read-only)")
    imported = property(lambda self: self._imported,
                        doc="This rule's imported status. See"
                            " :ref:`RefObjectModelRulesImported` for"
                            " more info.  (Read-only)")
    active   = property(lambda self: self._active,
                        doc="This rule's active state.  (Read-only)")
    enabled  = property(lambda self: self._enabled,
                        doc="This rule's enabled state. An enabled rule"
                            " is active when its context matches, a"
                            " disabled rule is never active regardless"
                            " of context.  (Read-only)")

    def enable(self):
        """
            Enable this rule so that it is active to receive
            recognitions when its context matches.

        """
        self._enabled = True
        self.activate()

    def disable(self):
        """
            Disable this rule so that it is never active to
            receive recognitions, regardless of whether its context
            matches or not.

        """
        self._enabled = False
        if self._active:
            self.deactivate()

    def _get_grammar(self):
        return self._grammar
    def _set_grammar(self, grammar):
        self._grammar = grammar
        # if self._grammar is None:
        #     self._grammar = grammar
        # elif grammar is None:
        #     self._grammar = None
        # elif grammar != self._grammar:
        #     self._log.error("rule: %s" % self)
        #     raise TypeError("The grammar object a Dragonfly rule"
        #         " cannot be changed after it has been set (%s != %s)."
        #         % (grammar, self._grammar))

    grammar = property(_get_grammar, _set_grammar,
                       doc="This rule's grammar object.  (Set once)")

    def set_context(self, context):
        """
            Set the context for this rule, under which it will be active and
            receive recognitions if it is also enabled and its grammar is
            active.

            Use of this method overwrites any previous context.

            Contexts can be modified at any time, but will only be checked
            when :meth:`process_begin` is called.

            :param context: context within which to be active.  If *None*,
                the rule will be active when its grammar is.
            :type context: Context|None
        """
        if not (isinstance(context, Context) or context is None):
            raise TypeError("context must be either a Context object or "
                            "None")
        self._context = context

    context = property(lambda self: self._context,
                       doc="This rule's context, under which it will be "
                           "active and receive recognitions if it is also "
                           "enabled and its grammar is active.")
    #-----------------------------------------------------------------------
    # Internal methods for controlling a rules active state.

    def process_begin(self, executable, title, handle):
        """
            Start of phrase callback.

            This method is called when the speech recognition
            engine detects that the user has begun to speak a
            phrase.  It is called by the rule's containing grammar
            if the grammar and this rule are active.

            The default implementation of this method checks
            whether this rule's context matches, and if it does
            this method calls
            :meth:`._process_begin`.

            Arguments:
             - *executable* -- the full path to the module whose
               window is currently in the foreground
             - *title* -- window title of the foreground window
             - *handle* -- window handle to the foreground window

        """
        if not self.grammar:
            raise GrammarError("A Dragonfly rule cannot be processed "
                               "before it is bound to a grammar.")
        if not self._enabled:
            if self._active:
                self.deactivate()
            return
        if self._context:
            if self._context.matches(executable, title, handle):
                if not self._active:
                    self.activate()
                self._process_begin()
            else:
                if self._active:
                    self.deactivate()
        else:
            if not self._active:
                self.activate()
            self._process_begin()

    def activate(self, force=False):
        if not self._grammar:
            raise GrammarError("A Dragonfly rule cannot be activated "
                               "before it is bound to a grammar.")
        if not self._enabled:
            if self._active:
                self.deactivate()
            return
        if not self._active or force:
            self._grammar.activate_rule(self)
            self._active = True

    def deactivate(self):
        if not self._grammar:
            raise GrammarError("A Dragonfly rule cannot be deactivated "
                               "before it is bound to a grammar.")
        if self._active:
            try:
                self._grammar.deactivate_rule(self)
            except Exception as e:
                self._log.warning("Failed to deactivate rule: %s (%s)",
                                  self, e)
            self._active = False

    #-----------------------------------------------------------------------
    # Compilation related methods.

    def gstring(self):
        s = "<" + self.name + ">"
        if self.imported: return s + " imported;"
        if self.exported: s += " exported"
        s += " = " + self.element.gstring() + ";"
        return s

    def dependencies(self, memo):
        if self._name in memo:
            return []
        memo.add(self._name)
        if self._element:
            return self._element.dependencies(memo)
        else:
            return []

    #-----------------------------------------------------------------------
    # Methods for decoding and evaluating recognitions.

    def decode(self, state):
        state.decode_attempt(self)

        for result in self._element.decode(state):
            state.decode_success(self)
            yield state
            state.decode_retry(self)

        state.decode_failure(self)

    def value(self, node):
        """
            Start of phrase callback.

            This method is called to obtain the semantic value associated
            with a particular recognition.  It could be called from
            another rule's :meth:`.value` if
            that rule references this rule.  If also be called from this
            rule's :meth:`.process_recognition`
            if that method has been overridden to do so in a derived
            class.

            The default implementation of this method returns the value of
            this rule's root element.

            .. note::

               This is generally the method which developers should
               override in derived rule classes to change the default
               semantic value of a recognized rule.

        """
        return node.children[0].value()

    #-----------------------------------------------------------------------
    # Methods for processing before-and-after utterances.

    def _process_begin(self):
        """
            Start of phrase detection callback.

            This method is called when the speech recognition
            engine detects that the user has begun to speak a
            phrase.  It is called by this rule's
            :meth:`.process_begin`
            after some context checks.

            The default implementation of this method does nothing.

            .. note::

               This is generally the method which developers should
               override in derived rule classes to give them custom
               functionality when the start of a phrase is detected.

        """

    def process_results(self, data):
        pass

    def process_recognition(self, node):
        """
            Rule recognition callback.

            This method is called when the user has spoken words matching
            this rule's contents.  This method is called only once for
            each recognition, and only for the matching top-level rule.

            The default implementation of this method does nothing.

            .. note::

               This is generally the method which developers should
               override in derived rule classes to give them custom
               functionality when a top-level rule is recognized.

        """


class ImportedRule(Rule):

    def __init__(self, name):
        Rule.__init__(self)
        self._name = name
        self._imported = True
        self._exported = False
        self._active = False
        self._grammar = None

    #-----------------------------------------------------------------------
    # Compilation related methods.

    def dependencies(self, memo):
        return ()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

from dragonfly import BasicRule, Repetition, Alternative, Literal, Function
from dragonfly.test import RuleTestCase


class TestBasicRule(RuleTestCase):

    def test_basic_rule(self):
        """ Verify that BasicRules can be loaded and recognized correctly.
        """
        test = []
        func = lambda x: test.append(x)

        # Test using BasicRule directly.
        rule = BasicRule(element=Repetition(
            Alternative((
                Literal("test one", value=Function(lambda: func(1))),
                Literal("test two", value=Function(lambda: func(2))),
                Literal("test three", value=Function(lambda: func(3))),
            )),
            1, 5
        ))
        self.add_rule(rule)
        self.recognize("test one test two test three".split())
        assert test == [1, 2, 3], "BasicRule was not processed correctly"

        # Remove the rule and clear the test list.
        self.grammar.remove_rule(rule)
        del test[:]

        # Test using a sub-class of BasicRule.
        class MyBasicRule(BasicRule):
            element = Repetition(
                Alternative((
                    Literal("test one", value=Function(lambda: func(1))),
                    Literal("test two", value=Function(lambda: func(2))),
                    Literal("test three", value=Function(lambda: func(3))),
                )),
                1, 5
            )
        self.add_rule(MyBasicRule())
        self.recognize("test one test two test three".split())
        assert test == [1, 2, 3], "BasicRule was not processed correctly"



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Test suite for German language Integer and Digits classes
============================================================================

"""

from dragonfly.test.infrastructure      import RecognitionFailure
from dragonfly.test.element_testcase    import ElementTestCase
from dragonfly.language.base.integer    import Integer
from dragonfly.language.de.number       import IntegerContent


#---------------------------------------------------------------------------

class GermanIntegerTestCase(ElementTestCase):
    """ Verify various German integers between 0 and 10**12. """
    def _build_element(self):
        return Integer(content=IntegerContent, min=0, max=10**12 - 1)
    input_output = [
        # Test a range of non-compound integer words.
        ("null",                                                        0),
        ("eins",                                                        1),
        ("ein",                                                         1),
        ("zwei",                                                        2),
        ("drei",                                                        3),
        ("vier",                                                        4),
        ("fuenf",                                                       5),
        ("sechs",                                                       6),
        ("sieben",                                                      7),
        ("acht",                                                        8),
        ("neun",                                                        9),
        ("zehn",                                                       10),
        ("elf",                                                        11),
        ("zwoelf",                                                     12),
        ("dreizehn",                                                   13),
        ("vierzehn",                                                   14),
        ("fuenfzehn",                                                  15),
        ("sechzehn",                                                   16),
        ("siebzehn",                                                   17),
        ("achtzehn",                                                   18),
        ("neunzehn",                                                   19),
        ("zwanzig",                                                    20),
        ("ein und zwanzig",                                            21),
        ("zwei und zwanzig",                                           22),
        ("drei und zwanzig",                                           23),
        ("ein und dreissig",                                           31),
        ("vier und dreissig",                                          34),
        ("fuenf und vierzig",                                          45),
        ("sechs und fuenfzig",                                         56),
        ("sieben und sechzig",                                         67),
        ("acht und siebzig",                                           78),
        ("neun und achtzig",                                           89),
        ("hundert",                                                   100),
        ("ein hundert",                                               100),
        ("ein hundert eins",                                          101),
        ("ein hundert ein",                            RecognitionFailure),
        ("ein hundert drei und zwanzig",                              123),
        ("drei hundert neun und achtzig",                             389),
        ("ein tausend eins",                                         1001),
        ("ein tausend ein",                            RecognitionFailure),
        ("ein tausend ein hundert ein und dreissig",                 1131),
        ("ein und dreissig tausend ein hundert ein und dreissig",   31131),
        ("ein hundert ein und dreissig tausend ein hundert ein "
         "und dreissig",                                           131131),
        ("million",                                               1000000),
        ("millionen",                                  RecognitionFailure),
        ("eine million eins",                                     1000001),
        ("eine million ein",                           RecognitionFailure),
        ("eine million zwei tausend eins",                        1002001),
        ("eine million zwei tausend ein",              RecognitionFailure),
        ("eine million ein hundert ein und dreissig tausend ein "
         "hundert ein und dreissig",                              1131131),
        ("zwei millionen",                                        2000000),
        ("ein hundert eins millionen",                          101000000),
        ("ein hundert ein millionen",                  RecognitionFailure),

        # Test a range of compound integer words.
        ("einundzwanzig",                                              21),
        ("zweiundzwanzig",                                             22),
        ("dreiundzwanzig",                                             23),
        ("einunddreissig",                                             31),
        ("vierunddreissig",                                            34),
        ("fuenfundvierzig",                                            45),
        ("sechsundfuenfzig",                                           56),
        ("siebenundsechzig",                                           67),
        ("achtundsiebzig",                                             78),
        ("neunundachtzig",                                             89),
        ("ein hundert dreiundzwanzig",                                123),
        ("ein hundert einunddreissig",                                131),
        ("drei hundert neunundachtzig",                               389),
        ("ein tausend ein hundert einunddreissig",                   1131),
        ("einunddreissig tausend ein hundert einunddreissig",       31131),
        ("ein hundert einunddreissig tausend ein hundert "
         "einunddreissig",                                         131131),
        ("eine million ein hundert einunddreissig tausend ein "
         "hundert einunddreissig",                                1131131),
    ]



"""This module contains any classes or functions which are used by multiple OS
accessibility controller implementations.
"""


class AccessibilityError(Exception):
    """Base class for checked exceptions."""


class UnsupportedSelectionError(AccessibilityError):
    """The built-in selection API is spotty when selecting across
    multiple objects. In Firefox you get results that look right, but the
    selection doesn't behave like a normal selection. Chrome simply does not
    support selections across multiple objects.
    """



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Mouse action
============================================================================

This section describes the :class:`Mouse` action object.  This type of
action is used for controlling the mouse cursor and clicking mouse
button.

Below you'll find some simple examples of :class:`Mouse` usage, followed
by a detailed description of the available mouse events.


Example mouse actions
............................................................................

The following code moves the mouse cursor to the center of the foreground
window (``(0.5, 0.5)``) and then clicks the left mouse button once
(``left``)::

    # Parentheses ("(...)") give foreground-window-relative locations.
    # Fractional locations ("0.5", "0.9") denote a location relative to
    #  the window or desktop, where "0.0, 0.0" is the top-left corner
    #  and "1.0, 1.0" is the bottom-right corner.
    action = Mouse("(0.5, 0.5), left")
    action.execute()

The line below moves the mouse cursor to 100 pixels left of the primary
monitor's left edge (if possible) and 250 pixels down from its top edge
(``[-100, 250]``), and then double clicks the right mouse button
(``right:2``)::

    # Square brackets ("[...]") give desktop-relative locations.
    # Integer locations ("1", "100", etc.) denote numbers of pixels.
    # Negative numbers ("-100") are counted from the left-edge of the
    #  primary monitor. They are used to access monitors above or to the
    #  left of the primary monitor.
    Mouse("[-100, 250], right:2").execute()

The following command drags the mouse from the top right corner of the
foreground window (``(0.9, 10), left:down``) to the bottom left corner
(``(25, -0.1), left:up``)::

    Mouse("(0.9, 10), left:down, (25, -0.1), left:up").execute()

The code below moves the mouse cursor 25 pixels right and 25 pixels up
(``<25, -25>``)::

    # Angle brackets ("<...>") move the cursor from its current position
    #  by the given number of pixels.
    Mouse("<25, -25>").execute()


Mouse specification format
............................................................................

The *spec* argument passed to the :class:`Mouse` constructor specifies
which mouse events will be emulated.  It is a string consisting of one or
more comma-separated elements.  Each of these elements has one of the
following possible formats:

Mouse movement actions:

 - move the cursor relative to the top-left corner of the desktop monitor
   containing coordinates ``[0, 0]`` (i.e. the primary monitor):
   ``[`` *number* ``,`` *number* ``]``
 - move the cursor relative to the foreground window:
   ``(`` *number* ``,`` *number* ``)``
 - move the cursor relative to its current position:
   ``<`` *pixels* ``,`` *pixels* ``>``

In the above specifications, the *number* and *pixels* have the
following meanings:

 - *number* -- can specify a number of pixels or a fraction of
   the reference window or desktop.  For example:

    - ``(10, 10)`` -- 10 pixels to the right and down from the
      foreground window's left-top corner
    - ``(0.5, 0.5)`` -- center of the foreground window

 - *pixels* -- specifies the number of pixels

Mouse button-press action:
   *keyname* [``:`` *repeat*] [``/`` *pause*]

 - *keyname* -- Specifies which mouse button to click:

    - ``left`` -- left mouse button key
    - ``middle`` -- middle mouse button key
    - ``right`` -- right mouse button key
    - ``four`` -- fourth mouse button key
    - ``five`` -- fifth mouse button key
    - ``wheelup`` -- mouse wheel up
    - ``stepup`` -- mouse wheel up 1/3
    - ``wheeldown`` -- mouse wheel down
    - ``stepdown`` -- mouse wheel down 1/3
    - ``wheelright`` -- mouse wheel right
    - ``stepright`` -- mouse wheel right 1/3
    - ``wheelleft`` -- mouse wheel left
    - ``stepleft`` -- mouse wheel left 1/3

 - *repeat* -- Specifies how many times the button should be clicked:

    - ``0`` -- don't click the button, this is a no-op
    - ``1`` -- normal button click
    - ``2`` -- double-click
    - ``3`` -- triple-click

 - *pause* --
   Specifies how long to pause *after* clicking the button.  The value
   should be an integer giving in hundredths of a second.  For example,
   ``/100`` would mean one second, and ``/50`` half a second.

Mouse button-hold or button-release action:
   *keyname* ``:`` *hold-or-release* [``/`` *pause*]

 - *keyname* -- Specifies which mouse button to click; same as above.

 - *hold-or-release* --
   Specified whether the button will be held down or released:

    - ``down`` -- hold the button down
    - ``up`` -- release the button

 - *pause* --
   Specifies how long to pause *after* clicking the button; same as above.


Mouse class reference
............................................................................

"""

# pylint: disable=R0201
# Suppress warnings about handler functions defined as Mouse methods.

from .action_base       import DynStrActionBase, ActionError
from ..windows.window   import Window

from .mouse import (ButtonEvent, PauseEvent, MoveRelativeEvent,
                    MoveScreenEvent, MoveWindowEvent, PLATFORM_BUTTON_FLAGS,
                    PLATFORM_WHEEL_FLAGS)

# Imported for backwards-compatibility: these functions used to live here.
from .mouse import get_cursor_position, set_cursor_position


class Mouse(DynStrActionBase):
    """ Action that sends mouse events. """

    def __init__(self, spec=None, static=False):
        """
            Arguments:
             - *spec* (*str*) -- the mouse actions to execute
             - *static* (boolean) --
               if *True*, do not dynamically interpret *spec*
               when executing this action

        """
        DynStrActionBase.__init__(self, spec=spec, static=static)

    def _parse_spec(self, spec):
        """ Convert the given *spec* to keyboard events. """
        events = []
        parts = self._split_parts(spec)
        for part in parts:
            handled = False
            for handler in self._handlers:
                try:
                    if handler(self, part, events):
                        handled = True
                        break
                except Exception:
                    continue
            if not handled:
                raise ActionError("Invalid mouse spec: %r (in %r)"
                                  % (part, spec))
        return events

    def _parse_position_pair(self, spec):
        parts = spec.split(",")
        if len(parts) != 2:
            raise ValueError("Invalid position pair spec: %r" % spec)

        h_origin, h_value = self._parse_position(parts[0])
        v_origin, v_value = self._parse_position(parts[1])

        return (h_origin, h_value, v_origin, v_value)

    def _process_window_position(self, spec, events):
        if not spec.startswith("(") or not spec.endswith(")"):
            return False
        h_origin, h_value, v_origin, v_value = self._parse_position_pair(spec[1:-1])
        event = MoveWindowEvent(h_origin, h_value, v_origin, v_value)
        events.append(event)
        return True

    def _process_screen_position(self, spec, events):
        if not spec.startswith("[") or not spec.endswith("]"):
            return False
        _, h_value, _, v_value = self._parse_position_pair(spec[1:-1])
        event = MoveScreenEvent(True, h_value, True, v_value)
        events.append(event)
        return True

    def _process_relative_position(self, spec, events):
        if not spec.startswith("<") or not spec.endswith(">"):
            return False
        parts = spec[1:-1].split(",")
        if len(parts) != 2:
            return False
        horizontal = int(parts[0])
        vertical   = int(parts[1])
        event = MoveRelativeEvent(horizontal, vertical)
        events.append(event)
        return True

    _button_flags = PLATFORM_BUTTON_FLAGS
    _wheel_flags = PLATFORM_WHEEL_FLAGS

    def _process_button(self, spec, events):
        parts = spec.split(":", 1)
        button = parts[0].strip()
        if len(parts) == 1:  special = 1
        else:                special = parts[1].strip()

        if button in self._button_flags:
            flag_down, flag_up = self._button_flags[button]

            if special == "down":
                event = ButtonEvent(flag_down)
            elif special == "up":
                event = ButtonEvent(flag_up)
            else:
                try:
                    repeat = int(special)
                except ValueError:
                    return False
                flag_series = (flag_down, flag_up) * repeat
                event = ButtonEvent(*flag_series)
        elif button in self._wheel_flags:
            flag = self._wheel_flags[button]
            try:
                repeat = int(special)
            except ValueError:
                return False
            flag = (flag[0], repeat * flag[1])
            event = ButtonEvent(flag)
        else:
            return False

        events.append(event)
        return True

    def _process_pause(self, spec, events):
        if not spec.startswith("/"):
            return False
        interval = float(spec[1:]) / 100
        event = PauseEvent(interval)
        events.append(event)
        return True

    _handlers = [
        _process_window_position,
        _process_screen_position,
        _process_relative_position,
        _process_button,
        _process_pause,
    ]

    def _parse_position(self, spec):
        spec = spec.strip()

        if spec.startswith("-"):  origin = False
        else:                     origin = True

        if spec.find(".") != -1:  value = float(spec)
        else:                     value = int(spec)

        return (origin, value)

    def _split_parts(self, spec):

        delimiters = ["()", "[]", "<>"]

        parts = spec.split(",")
        items = []
        while parts:
            part = parts.pop(0).strip()

            found_delimiter = False
            for begin, end in delimiters:
                if begin not in part:
                    continue

                item = [part]
                while parts:
                    part = parts.pop(0)
                    item.append(part)
                    if end in part:
                        break
                if end not in part:
                    raise Exception("No closing delimiter found for %r" % item[0])

                new_item = ",".join(item)
                found_delimiter = True
                break

            if not found_delimiter:
                new_item = part

            if "/" in new_item:
                before, after = new_item.split("/", 1)
                items.append(before.strip())
                items.append("/" + after.strip())
            else:
                items.append(new_item)

        return items


    #-----------------------------------------------------------------------

    def _execute_events(self, events):
        """ Send events. """
        window = Window.get_foreground()
        for event in events:
            event.execute(window)

    def __str__(self):
        return '<{}>'.format(self._spec)



#
# This file is a command-module for Dragonfly.
# (c) Copyright 2008 by Christo Butcher
# Licensed under the LGPL, see <http://www.gnu.org/licenses/>
#

"""
Command-module loader for natlink
=================================

This script can be used to look Dragonfly command-modules
for use with Dragon NaturallySpeaking.  It scans the
directory it's in and loads any ``_*.py`` it finds.

Some notes
----------

- Modules loaded by this script will **not** be re-loaded
  automatically.

- Messages printed from loaded modules will **not** show up
  in natlink's messages window because this script uses a
  separate process.

- Modules loaded normally via natlinkmain will still work
  and be recognized normally.

- This script will open a small dialog window via the
  ``natlink.waitForSpeech()`` function and will terminate
  when you click 'Close'.

- One use for this script is to control admin applications with
  Dragon. This can be done by simply running the script as the
  administrator with Dragon running and dragonfly's dependencies
  installed.

"""

import os.path
import sys

import six

from dragonfly import get_engine
from dragonfly.loader import CommandModuleDirectory
from dragonfly.log import setup_log

#---------------------------------------------------------------------------
# Set up basic logging.

setup_log()
# logging.getLogger("compound.parse").setLevel(logging.INFO)


#---------------------------------------------------------------------------
# Main event driving loop.

def main():
    try:
        path = os.path.dirname(__file__)
    except NameError:
        # The "__file__" name is not always available, for example
        #  when this module is run from PythonWin.  In this case we
        #  simply use the current working directory.
        path = os.getcwd()
        __file__ = os.path.join(path, "dfly-loader-natlink.py")

    # Initialize and connect the engine.
    engine = get_engine("natlink")
    engine.connect()

    # Load grammars.
    directory = CommandModuleDirectory(path, excludes=[__file__])
    directory.load()

    # Define recognition callback functions.
    def on_begin():
        print("Speech start detected.")

    def on_recognition(words):
        message = u"Recognized: %s" % u" ".join(words)

        # This only seems to be an issue with Python 2.7 on Windows.
        if six.PY2:
            encoding = sys.stdout.encoding or "ascii"
            message = message.encode(encoding, errors='replace')
        print(message)

    def on_failure():
        print("Sorry, what was that?")

    # Recognize from Dragon in a loop (opens a dialogue window).
    try:
        engine.do_recognition(on_begin, on_recognition, on_failure)
    except KeyboardInterrupt:
        pass

    # Disconnect after the dialogue is closed.
    engine.disconnect()


if __name__ == "__main__":
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
ActionBase base class
============================================================================

"""

from functools import reduce
from locale import getpreferredencoding
import logging

from six import PY2, integer_types, text_type


#---------------------------------------------------------------------------


class ActionError(Exception):
    pass


#---------------------------------------------------------------------------

class ActionBase(object):
    """
        Base class for Dragonfly's action classes.

    """

    _log_init = logging.getLogger("action.init")
    _log_exec = logging.getLogger("action.exec")
    _log = logging.getLogger("action")

    #-----------------------------------------------------------------------
    # Initialization and aggregation methods.

    def __init__(self):
        self._str = ""

    def __repr__(self):
        if PY2:
            return self.__unicode__().encode(getpreferredencoding())
        else:
            return self.__unicode__()

    def __unicode__(self):
        return u"%s(%s)" % (self.__class__.__name__, self._str)

    def __add__(self, other):
        return ActionSeries(self, other)

    def __iadd__(self, other):
        return ActionSeries(self, other)

    def __or__(self, other):
        return UnsafeActionSeries(self, other)

    def __ior__(self, other):
        return UnsafeActionSeries(self, other)

    def __mul__(self, factor):
        return ActionRepetition(self, factor)

    def __imul__(self, factor):
        return ActionRepetition(self, factor)

    #-----------------------------------------------------------------------
    # Execution methods.

    def bind(self, data=None):
        return BoundAction(self, data)

    def copy_bind(self, data=None):
        return BoundAction(self, data)

    def execute(self, data=None):
        self._log_exec.debug("Executing action: %s (%s)", self, data)
        try:
            if self._execute(data) is False:
                raise ActionError(str(self))
        except ActionError as e:
            self._log_exec.error("Execution failed: %s", e)
            return False
        except Exception as e:
            self._log_exec.exception("Execution of %s failed due to "
                                     "exception: %s", self, e)
            return False
        return True

    def _execute(self, data=None):
        """ Virtual method. """


#---------------------------------------------------------------------------

class DynStrActionBase(ActionBase):

    # pylint: disable=E1111,R1710
    # Suppress warnings about return statements in some methods.

    #-----------------------------------------------------------------------
    # Initialization methods.

    def __init__(self, spec=None, static=False):
        ActionBase.__init__(self)
        self.initialize(spec, static)

    def initialize(self, spec=None, static=False):
        self._spec = spec
        self._static = False
        self._events = None
        if spec is None: return

        if static or spec.find("%") == -1:
            self._static = True
            self._events = self._parse_spec(spec)
        else:
            self._static = False
            self._events = None

        self._str = "%r" % spec
        if not self._static:
            self._str += ", dynamic"

    def _parse_spec(self, spec):
        """ Virtual method. """

    #-----------------------------------------------------------------------
    # Execution methods.

    def _execute(self, data=None):
        if self._static:
            # If static, the events have already been parsed by the
            #  initialize() method.
            self._execute_events(self._events)

        else:
            # If not static, now is the time to build the dynamic spec,
            #  parse it, and execute the events.

            if not data:
                spec = self._spec
            else:
                try:
                    spec = self._spec % data
                except KeyError:
                    self._log_exec.error("%s: Spec %r doesn't match data "
                                         "%r.", self, self._spec, data)
                    return False

            self._log_exec.debug("%s: Parsing dynamic spec: %r",
                                 self, spec)
            events = self._parse_spec(spec)
            self._execute_events(events)

    def _execute_events(self, events):
        """ Virtual method. """


#---------------------------------------------------------------------------

class BoundAction(ActionBase):

    #-----------------------------------------------------------------------
    # Initialization methods.

    def __init__(self, action, data):
        ActionBase.__init__(self)
        self._action = action
        self._data = data
        self._str = "%r, %r" % (action, data)

    #-----------------------------------------------------------------------
    # Execution methods.

    def execute(self, data=None):
        if not data:
            data = {}
        if self._data:
            data = dict(data)
            data.update(self._data)

        self._action.execute(data)


#---------------------------------------------------------------------------

class ActionSeries(ActionBase):

    #-----------------------------------------------------------------------
    # Initialization methods.

    #: Whether to stop executing if an action in the series fails.
    stop_on_failures = True

    def __init__(self, *actions):
        ActionBase.__init__(self)
        self._actions = list(actions)
        self._set_str()

    def _set_str(self):
        # Use a flat list of the series actions for a more readable
        # string representation.
        self._str = u", ".join(text_type(a)
                               for a in self.flat_action_list())

    def flat_action_list(self):
        # Get a flattened list of the series actions.
        result = []
        for action in self._actions:
            if isinstance(action, ActionSeries) and action.stop_on_failures:
                result.extend(action.flat_action_list())
            else:
                result.append(action)
        return result

    def append(self, other):
        assert isinstance(other, ActionBase)
        self._actions.append(other)
        self._set_str()

    def __iadd__(self, other):
        self.append(other)
        return self

    def __ior__(self, other):
        self.append(other)
        return self

    #-----------------------------------------------------------------------
    # Execution methods.

    def _execute(self, data=None):
        # Use a flat list of the series actions for more sensible sequence
        # termination and logging if an error occurs during execution.
        for action in self.flat_action_list():
            if action.execute(data) is False and self.stop_on_failures:
                return False
        return True

    def execute(self, data=None):
        # Override execute() to discard the return value.
        ActionBase.execute(self, data)

    def __str__(self):
        return reduce((lambda x, y: "{}+{}".format(x, y)), self._actions)


class UnsafeActionSeries(ActionSeries):
    stop_on_failures = False

    def execute(self, data=None):
        for action in self._actions:
            action.execute(data)

    def __str__(self):
        return reduce((lambda x, y: "{}|{}".format(x, y)), self._actions)


#---------------------------------------------------------------------------

class ActionRepetition(ActionBase):

    #-----------------------------------------------------------------------
    # Initialization methods.

    def __init__(self, action, factor):
        ActionBase.__init__(self)
        self._action = action
        self._factor = factor
        self._str = "%s, %s" % (action, factor)

        if not isinstance(factor, (int, Repeat)):
            raise TypeError("Invalid multiplier type: %r"
                            " (must be an int or a Repeat object)" % factor)

    #-----------------------------------------------------------------------
    # Execution methods.

    def execute(self, data=None):
        if isinstance(self._factor, integer_types):
            repeat = self._factor
        elif isinstance(self._factor, Repeat):
            repeat = self._factor.factor(data)
        else:
            raise ActionError("Invalid repeat factor: %r" % (self._factor,))

        for _ in range(repeat):
            if self._action.execute(data) is False:
                raise ActionError(str(self))

    def __str__(self):
        return '{{{}}}{}'.format(self._action, self._factor)


#---------------------------------------------------------------------------

class Repeat(object):
    # pylint: disable=line-too-long

    """
        Action repeat factor.

        Integer Repeat factors ignore any supply data::

            >>> integer = Repeat(count=3)
            >>> integer.factor()
            3
            >>> integer.factor({"foo": 4})  # Non-related data is ignored.
            3

        Integer Repeat factors can be specified with the ``*`` operator::

            >>> from dragonfly import Function
            >>> def func():
            ...     print("executing 'func'")
            ...
            >>> action = Function(func) * 3
            >>> action.execute()
            executing 'func'
            executing 'func'
            executing 'func'

        Named Repeat factors retrieved their factor-value from the
        supplied data::

            >>> named = Repeat("foo")
            >>> named.factor()
            Traceback (most recent call last):
              ...
            ActionError: No extra repeat factor found for name 'foo' ('NoneType' object is unsubscriptable)
            >>> named.factor({"foo": 4})
            4

        Repeat factors with both integer count and named extra values set
        combined (add) these together to determine their factor-value::

            >>> combined = Repeat(extra="foo", count=3)
            >>> combined.factor()
            Traceback (most recent call last):
              ...
            ActionError: No extra repeat factor found for name 'foo' ('NoneType' object is unsubscriptable)
            >>> combined.factor({"foo": 4}) # Combined factors 3 + 4 = 7.
            7

    """

    def __init__(self, extra=None, count=None):
        # Backward compatibility for swapped arguments
        # (#103)
        if isinstance(extra, integer_types):
            self._count = extra
            self._extra = count
        else:
            self._extra = extra
            self._count = count if count is not None else 0

    def factor(self, data=None):
        count = self._count
        if self._extra:
            try:
                additional = int(data[self._extra])
            except ValueError:
                raise ActionError("Repeat factor %r has invalid value %r"
                                  % (self._extra, data[self._extra]))
            except KeyError:
                raise ActionError("No extra repeat factor found for name %r"
                                  % (self._extra,))
            except Exception as e:
                raise ActionError("No extra repeat factor found for name %r"
                                  " (%s)" % (self._extra, e))
            count += additional
        return count

    def __str__(self):
        return '{}'.format(self._extra if self._extra else self._count)
